# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.packages.pipeline import RelengPackagePipeline
from tests import prepare_testenv
import os
import unittest

class TestPkgPipelineLicenses(unittest.TestCase):
    def test_pkg_pipeline_licenses_multiple(self):
        pkg_names = [
            'test-a',
            'test-b',
            'test-c',
        ]

        expected_licenses = [
            2, # mocked BSD + MIT
            None,
            1, # mocked GPL
        ]

        with prepare_testenv(template='licenses') as engine:
            pkgs = engine.pkgman.load(pkg_names)

            pipeline = RelengPackagePipeline(engine, engine.opts, {})
            for pkg in pkgs:
                if not pipeline.process(pkg):
                    break

            for pkg, expected in zip(pkg_names, expected_licenses):
                if expected is not None:
                    # a package with license information should be tracked
                    self.assertTrue(pkg in pipeline.license_files)

                    # verify we have expected license counts
                    licenses = pipeline.license_files[pkg]['files']
                    self.assertEqual(len(licenses), expected)

                    # verify that the license reference maps to a real file
                    for license_ in licenses:
                        self.assertTrue(os.path.exists(license_))
                else:
                    # packages without license data should not provide any
                    # information
                    self.assertFalse(pkg in pipeline.license_files)

    def test_pkg_pipeline_licenses_none(self):
        with prepare_testenv(template='licenses') as engine:
            pkgs = engine.pkgman.load(['test-b'])

            pipeline = RelengPackagePipeline(engine, engine.opts, {})
            for pkg in pkgs:
                if not pipeline.process(pkg):
                    break

            # verify we have no license information if packages have no license
            # information to provide
            self.assertFalse(pipeline.license_files)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.util.env import env_value
from releng_tool.util.env import extend_script_env
import os
import unittest

class TestUtilEnv(unittest.TestCase):
    def test_utilenv_extendscriptenv(self):
        env = {'a': 1}

        # can append new entries
        extend_script_env(env, {'b': 2})
        self.assertEqual(len(env.keys()), 2)
        self.assertIn('a', env)
        self.assertIn('b', env)
        self.assertEqual(env['a'], 1)
        self.assertEqual(env['b'], 2)

        # will override existing entries
        extend_script_env(env, {'b': 3})
        self.assertEqual(len(env.keys()), 2)
        self.assertIn('a', env)
        self.assertIn('b', env)
        self.assertEqual(env['a'], 1)
        self.assertEqual(env['b'], 3)

        # magic values are ignored
        extend_script_env(env, {'__magic__': 4})
        self.assertEqual(len(env.keys()), 2)

        # imported built-in functions are ignored
        extend_script_env(env,
            {'test': globals()['__builtins__']['hash']})
        self.assertEqual(len(env.keys()), 2)

        # imported functions are ignored
        extend_script_env(env,
            {'test': globals()['extend_script_env']})
        self.assertEqual(len(env.keys()), 2)

        # imported modules are ignored
        extend_script_env(env, {'test': globals()['unittest']})
        self.assertEqual(len(env.keys()), 2)

    def test_utilenv_env_value(self):
        test_env_key = 'RELENG_TOOL_UNIT_TEST_KEY'
        test_env_val = 'NEW_VALUE'

        # check a (should be) empty/unset environment variable
        value = os.environ.get(test_env_key)
        self.assertIsNone(value)

        value = env_value(test_env_key)
        self.assertIsNone(value)

        # configure an environment variable
        value = env_value(test_env_key, test_env_val)
        self.assertEqual(value, test_env_val)

        value = env_value(test_env_key)
        self.assertEqual(value, test_env_val)

        value = os.environ.get(test_env_key)
        self.assertEqual(value, test_env_val)

        # configure an empty environment variable
        value = env_value(test_env_key, '')
        self.assertEqual(value, '')

        value = env_value(test_env_key)
        self.assertEqual(value, '')

        value = os.environ.get(test_env_key)
        self.assertEqual(value, '')

        # remove the environment variable
        value = env_value(test_env_key, None)
        self.assertIsNone(value)

        value = env_value(test_env_key)
        self.assertIsNone(value)

        value = os.environ.get(test_env_key)
        self.assertIsNone(value)

        # retry removing the environment variable (no-op)
        value = env_value(test_env_key, None)
        self.assertIsNone(value)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.util.io import opt_file
from releng_tool.util.io import run_script
from releng_tool.util.log import verbose
import os

#: filename of the script to execute the building operation (if any)
BUILD_SCRIPT = 'build'

def build(opts):
    """
    support building project-defined scripts

    With provided build options (``RelengBuildOptions``), the build stage will
    be processed.

    Args:
        opts: build options

    Returns:
        ``True`` if the building stage is completed; ``False`` otherwise
    """

    assert opts
    build_dir = opts.build_dir
    def_dir = opts.def_dir
    env = opts.env

    build_script_filename = '{}-{}'.format(opts.name, BUILD_SCRIPT)
    build_script = os.path.join(def_dir, build_script_filename)
    build_script, build_script_exists = opt_file(build_script)
    if not build_script_exists:
        if (opts._skip_remote_scripts or
                'releng.disable_remote_scripts' in opts._quirks):
            return True

        build_script_filename = '{}-{}'.format('releng', BUILD_SCRIPT)
        build_script = os.path.join(build_dir, build_script_filename)
        build_script, build_script_exists = opt_file(build_script)
        if not build_script_exists:
            return True

    if not run_script(build_script, env, subject='build'):
        return False

    verbose('install script executed: ' + build_script)
    return True



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

import os

try:
    basestring  # pylint: disable=E0601
except NameError:
    basestring = str

try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence  # pylint: disable=W1512

def expand(obj, kv=None):
    """
    perform variable expansion on strings

    This expand utility method will attempt to expand variables in detected
    string types. For a detected string which contains substrings in the form of
    ``$value`` or ``${value}``, these substrings will be replaced with their
    respective key-value (if provided) or environment variable value. For
    substrings which do not have a matching variable value, the substrings will
    be replaced with an empty value. If a dictionary is provided, keys and
    values will be checked if they can be expanded on. If a list/set is
    provided, each value which be checked if it can be expanded on. If a
    dictionary key is expanded to match another key, a key-value pair can be
    dropped. If a set may result in a smaller set if expanded values result in
    duplicate entries.

    An example when using in the context of script helpers is as follows:

    .. code-block:: python

        import os
        ...

        os.environ['MY_ENV'] = 'my-environment-variable'
        value = releng_expand('$MY_ENV')
        print(value)
        # will output: my-environment-variable

    Args:
        obj: the object
        kv (optional): key-values pairs to use

    Returns:
        the expanded object
    """
    if isinstance(obj, basestring):
        try:
            idx = obj.index('$')
        except ValueError:
            return obj

        if kv:
            final_kv = dict(os.environ)
            final_kv.update(kv)
        else:
            final_kv = os.environ

        rv = obj[:idx]
        objlen = len(obj)

        while idx < objlen:
            c = obj[idx:idx + 1]
            if c == '$':
                nc = obj[idx + 1:idx + 2]
                if not nc or nc == '$' or nc.isspace():
                    rv += c
                    idx += 1
                elif nc == '{':
                    try:
                        eidx = obj.index('}', idx + 1)
                        var = obj[idx + 2:eidx]
                        if var in final_kv:
                            rv += final_kv[var]
                        idx = eidx
                    except ValueError:
                        rv += obj[idx:]
                        break
                else:
                    var = nc
                    idx += 1

                    nc = obj[idx + 1:idx + 2]
                    while nc and nc != '$' and (nc.isalnum() or nc == '_'):
                        var += nc
                        idx += 1
                        nc = obj[idx + 1:idx + 2]

                    if var in final_kv:
                        rv += final_kv[var]
            else:
                rv += c
            idx += 1
    elif isinstance(obj, dict):
        rv = {}
        for key, value in obj.items():
            rv[expand(key)] = expand(value)
    elif isinstance(obj, list):
        rv = []
        for value in obj:
            rv.append(expand(value))
    elif isinstance(obj, set):
        rv = set()
        for value in obj:
            rv.add(expand(value))
    else:
        rv = obj

    return rv

def interpret_dictionary_strings(obj):
    """
    interpret a dictionary of key-value strings from the provided object

    Attempts to interpret one or more key-value string pairs from a provided
    object. If a key-value string dictionary value is provided, it will be
    returned. In the case where an unexpected type is detected, this method will
    return ``None``.

    Args:
        obj: the object to interpret

    Returns:
        the dictionary; otherwise ``None``
    """
    rv = None

    if isinstance(obj, dict):
        rv = obj
        for key, value in obj.items():
            if not isinstance(key, basestring):
                rv = None
                break
            elif value is not None and not isinstance(value, basestring):
                rv = None
                break

    return rv

def interpret_string(obj):
    """
    interpret a string, if any, from the provided object

    Attempts to interpret string from a provided object. If a string value is
    provided, it will be returned. In the case where an unexpected type is
    detected, this method will return ``None``.

    Args:
        obj: the object to interpret

    Returns:
        the string; otherwise ``None``
    """
    rv = None

    if isinstance(obj, basestring):
        rv = obj

    return rv

def interpret_strings(obj):
    """
    interpret strings, if any, from the provided object

    Attempts to interpret one or more strings from a provided object. Returned
    will be an iterable containing one or more strings. If a string value is
    provided, it will be returned inside iterable container. If an iterable
    container is provided, the same container will be returned. In the case
    where an unexpected type is detected, this method will return ``None``.

    Args:
        obj: the object to interpret

    Returns:
        sequence of zero or more strings; otherwise ``None``
    """
    rv = None

    if isinstance(obj, Sequence):
        if isinstance(obj, basestring):
            rv = [obj]
        else:
            rv = obj
            for child in obj:
                if not isinstance(child, basestring):
                    rv = None
                    break

    return rv

def interpret_zero_to_one_strings(obj):
    """
    interpret a dictionary of zero-to-one strings from the provided object

    Attempts to interpret one or more zero-to-one strings from a provided
    object. A zero-to-one string is a string-based key which may or may not have
    an associated value assigned to it. If a key-value string dictionary value
    is provided, it will be returned. If the sequence of strings or a single
    string is provided, a dictionary will be populated with matching keys to
    provided names with empty string values. In the case where an unexpected
    type is detected, this method will return ``None``.

    Args:
        obj: the object to interpret

    Returns:
        the dictionary; otherwise ``None``
    """
    rv = None

    if isinstance(obj, dict):
        rv = obj
        for key, value in obj.items():
            if not isinstance(key, basestring):
                rv = None
                break
            elif value is not None and not isinstance(value, basestring):
                rv = None
                break
    elif isinstance(obj, Sequence):
        rv = {}
        if isinstance(obj, basestring):
            rv[obj] = ''
        else:
            for child in obj:
                if not isinstance(child, basestring):
                    rv = None
                    break
                rv[child] = ''

    return rv

def is_sequence_not_string(obj):
    """
    return whether or not the provided object is a non-string sequence

    Returns ``True`` if the provided ``obj`` is a sequence type but is also not
    a string; ``False`` otherwise.

    Args:
        obj: the object to interpret

    Returns:
        whether or not a non-string sequence
    """
    return isinstance(obj, Sequence) and not isinstance(obj, basestring)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.make import MAKE
from releng_tool.util.io import prepare_arguments
from releng_tool.util.io import prepare_definitions
from releng_tool.util.log import err
from releng_tool.util.string import expand

def install(opts):
    """
    support installation autotools projects

    With provided installation options (``RelengInstallOptions``), the
    installation stage will be processed.

    Args:
        opts: installation options

    Returns:
        ``True`` if the installation stage is completed; ``False`` otherwise
    """

    if not MAKE.exists():
        err('unable to install package; make is not installed')
        return False

    # definitions
    autotools_defs = {
    }
    if opts.install_defs:
        autotools_defs.update(expand(opts.install_defs))

    # default options
    autotools_opts = {
    }
    if opts.install_opts:
        autotools_opts.update(expand(opts.install_opts))

    # argument building
    autotools_args = [
    ]

    # If the provided package has not provided any installation options,
    # indicate that the install target will be run.
    if not opts.install_opts:
        autotools_args.append('install')

    autotools_args.extend(prepare_definitions(autotools_defs))
    autotools_args.extend(prepare_arguments(autotools_opts))

    # install to each destination
    env = expand(opts.install_env)
    for dest_dir in opts.dest_dirs:
        if not MAKE.execute(['DESTDIR=' + dest_dir] + autotools_args, env=env):
            err('failed to install autotools project: {}', opts.name)
            return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2020 releng-tool

from releng_tool.util.log import warn
from releng_tool.__main__ import main
import sys

if __name__ == '__main__':
    warn('(deprecated) releng module is deprecated over releng_tool')
    sys.exit(main())



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.util.io import opt_file
from releng_tool.util.io import run_script
from releng_tool.util.log import verbose
import os

#: filename of the script to execute the configuration operation (if any)
CONFIGURE_SCRIPT = 'configure'

def configure(opts):
    """
    support configuration project-defined scripts

    With provided configuration options (``RelengConfigureOptions``), the
    configuration stage will be processed.

    Args:
        opts: configuration options

    Returns:
        ``True`` if the configuration stage is completed; ``False`` otherwise
    """

    assert opts
    build_dir = opts.build_dir
    def_dir = opts.def_dir
    env = opts.env

    configure_script_filename = '{}-{}'.format(opts.name, CONFIGURE_SCRIPT)
    configure_script = os.path.join(def_dir, configure_script_filename)
    configure_script, configure_script_exists = opt_file(configure_script)
    if not configure_script_exists:
        if (opts._skip_remote_scripts or
                'releng.disable_remote_scripts' in opts._quirks):
            return True

        configure_script_filename = '{}-{}'.format('releng', CONFIGURE_SCRIPT)
        configure_script = os.path.join(build_dir, configure_script_filename)
        configure_script, configure_script_exists = opt_file(configure_script)
        if not configure_script_exists:
            return True

    if not run_script(configure_script, env, subject='configure'):
        return False

    verbose('install script executed: ' + configure_script)
    return True



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from contextlib import contextmanager
from difflib import unified_diff
from io import open
from releng_tool.engine import RelengEngine
from releng_tool.opts import RelengEngineOptions
from releng_tool.util.io import generate_temp_dir
from releng_tool.util.io import interim_working_dir
from releng_tool.util.io import path_copy
import os
import sys
import unittest

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

def compare_contents(first, second):
    """
    compare the contents of two files

    This utility method is used to compare the contents of two files. Both
    file's contents will be read and checked for any differences. If both
    files have the same contents, `None` will be returned; otherwise a
    string containing a unified diff string will be returned.

    Args:
        first: the first file
        second: the second file

    Returns:
        `None` if matching; otherwise a unified diff string
    """
    def strip_lines(lines):
        return [line.strip() + '\n' for line in lines]

    try:
        with open(first, mode='r', encoding='utf_8') as file:
            content1 = strip_lines(file.readlines())
    except IOError:
        return 'failed to load first file: ' + first

    try:
        with open(second, mode='r', encoding='utf_8') as file:
            content2 = strip_lines(file.readlines())
    except IOError:
        return 'failed to load second file: ' + second

    diff = unified_diff(content1, content2,
        fromfile=first, tofile=second, lineterm='\n')
    diff_content = ''.join(list(diff))
    if diff_content:
        return 'unexpected file differences\n{}'.format(diff_content)

    return None

@contextmanager
def prepare_testenv(config=None, template=None, args=None):
    """
    prepare an engine-ready environment for a test

    This utility method is used to provide an `RelengEngine` instance ready for
    execution on an interim working directory.

    Args:
        config (optional): dictionary of options to mock for arguments
        template (optional): the folder holding a template project to copy into
                              the prepared working directory
        args (optional): additional arguments to add to the "forwarded options"

    Yields:
        the engine
    """

    class MockArgs(object):
        def __getattr__(self, name):
            return self.name if name in self.__dict__ else None

    if config is None:
        config = {}

    with generate_temp_dir() as work_dir, interim_working_dir(work_dir):
        # force root directory to temporary directory; or configure all working
        # content based off the generated temporary directory
        if 'root_dir' not in config:
            config['root_dir'] = work_dir
        else:
            if 'cache_dir' not in config:
                config['cache_dir'] = os.path.join(work_dir, 'cache')
            if 'dl_dir' not in config:
                config['dl_dir'] = os.path.join(work_dir, 'dl')
            if 'out_dir' not in config:
                config['out_dir'] = os.path.join(work_dir, 'out')

        if template:
            templates_dir = os.path.join(find_test_base(), 'templates')
            template_dir = os.path.join(templates_dir, template)
            if not path_copy(template_dir, work_dir, critical=False):
                assert False, 'failed to setup template into workdir'

        # build arguments instance
        test_args = MockArgs()
        for k, v in config.items():
            setattr(test_args, k, v)

        # prepare engine options and build an engine instance
        opts = RelengEngineOptions(args=test_args, forward_args=args)
        engine = RelengEngine(opts)

        yield engine

@contextmanager
def prepare_workdir():
    """
    prepare a working directory for a test

    This utility method is used to provide a test a directory to store
    output files. This method will ensure the container directory is emptied
    before returning.

    Returns:
        the container directory
    """

    with generate_temp_dir() as work_dir:
        yield work_dir

@contextmanager
def redirect_stderr(new_target=None):
    """
    temporarily redirect stderr to another instance

    This call will temporarily redirect stderr to the provided instance
    until the end of the context, where the previous `stderr` target will be
    restored.

    Args:
        new_target (optional): the instance to map stderr to
    """

    if not new_target:
        new_target = StringIO()

    _ = sys.stderr
    try:
        sys.stderr = new_target
        yield new_target
    finally:
        sys.stderr = _

@contextmanager
def redirect_stdout(new_target=None):
    """
    temporarily redirect stdout to another instance

    This call will temporarily redirect stdout to the provided instance
    until the end of the context, where the previous `stdout` target will be
    restored.

    Args:
        new_target (optional): the instance to map stdout to
    """

    if not new_target:
        new_target = StringIO()

    _ = sys.stdout
    try:
        sys.stdout = new_target
        yield new_target
    finally:
        sys.stdout = _

def run_testenv(config=None, template=None, args=None):
    """
    execute an engine instance with provide environment options for a test

    This utility method is used to invoke an `RelengEngine` instance which is
    prepared based off the provided configuration options and base template (if
    any) to copy.

    Args:
        config (optional): dictionary of options to mock for arguments
        template (optional): the folder holding a template project to copy into
                              the prepared working directory
        args (optional): additional arguments to add to the "forwarded options"

    Returns:
        the engine's return code
    """

    with prepare_testenv(config=config, template=template, args=args) as engine:
        return engine.run()

def find_test_base():
    """
    return the absolute path of the test base directory

    A utility call to return the absolute path of the "tests" directory for this
    implementation. This is to support interpreters (i.e. Python 2.7) which do
    not provide an absolute path via the `__file__` variable.

    Returns:
        the path
    """

    test_base = os.path.dirname(os.path.realpath(__file__))

    def templates_exist(test_base):
        return os.path.exists(os.path.join(test_base, 'templates'))

    if not templates_exist(test_base):
        test_base = os.path.dirname(os.path.abspath(sys.argv[0]))
        if not templates_exist(test_base):
            # python 2.7 may not always be able to find the test directory at
            # all stages of test; try to rely on a provided tox ini directory if
            # running in a tox environment
            if sys.version_info < (3, 0) and 'TOX_INI_DIR' in os.environ:
                root_dir = os.environ['TOX_INI_DIR']
                test_base = os.path.join(root_dir, 'tests')

            if not templates_exist(test_base):
                raise RuntimeError('unable to find test base directory')

    return test_base

class RelengToolTestSuite(unittest.TestSuite):
    def run(self, result, debug=False):
        """
        a releng-tool helper test suite

        Provides a `unittest.TestSuite` which will ensure stdout is flushed
        after the execution of tests. This is to help ensure all stdout content
        from the test is output to the stream before the unittest framework
        outputs a test result summary which may be output to stderr.

        See `unittest.TestSuite.run()` for more details.

        Args:
            result: the test result object to populate
            debug (optional): debug flag to ignore error collection

        Returns:
            the test result object
        """
        rv = unittest.TestSuite.run(self, result, debug)
        sys.stdout.flush()
        return rv



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.util.io import _execute
from releng_tool.util.io import execute
from releng_tool.util.log import debug
from releng_tool.util.log import err
from releng_tool.util.string import is_sequence_not_string
import os
import re

class RelengTool:
    """
    a host tool

    Provides a series of host tools methods to assist in validating the
    existence of a host tool as well as the execution of a host tool.

    Attributes:
        detected: tracking whether or not a tool is available on the host system

    Args:
        tool: the file name of the tool
        exists_args (optional): argument value to check for existence (no-op)
        env_sanitize (optional): environment variables to sanitize
        env_include (optional): environment variables to always include
    """
    detected = {}

    def __init__(self, tool, exists_args=None, env_sanitize=None,
            env_include=None):
        self.include = env_include
        self.sanitize = env_sanitize

        # allow a system to override a host tool path
        override_tool_key = 'RELENG_' + re.sub(r'[^A-Z0-9]', '', tool.upper())
        if override_tool_key in os.environ:
            self.tool = os.environ[override_tool_key]
        else:
            self.tool = tool

        if exists_args is not None:
            self.exists_args = exists_args
        else:
            self.exists_args = ['--version']

    def execute(self, args=None, cwd=None, quiet=False, env=None, poll=False,
            capture=None):
        """
        execute the host tool with the provided arguments (if any)

        Runs the host tool described by ``args`` until completion.

        Args:
            args (optional): the list of arguments for the tool
            cwd (optional): working directory to use
            quiet (optional): whether or not to suppress output
            env (optional): environment variables to include
            poll (optional): force polling stdin/stdout for output data
            capture (optional): list to capture output into

        Returns:
            ``True`` if the execution has completed with no error; ``False`` if
            the execution has failed
        """

        rv = self._execute(args=args, cwd=cwd, quiet=quiet, env=env, poll=poll,
            capture=capture)
        return (rv == 0)

    def execute_rv(self, *args, **kwargs):
        """
        execute the host tool with the provided arguments (if any)

        Runs the host tool described by ``args`` until completion.

        Args:
            *args (optional): arguments to add to the command
            **cwd: working directory to use
            **env: environment variables to include

        Returns:
            the return code of the execution request
        """

        out = []
        rv = self._execute(list(args),
            cwd=kwargs.get('cwd'),
            env=kwargs.get('env'),
            capture=out, quiet=True)
        return rv, '\n'.join(out)

    def _execute(self, args=None, cwd=None, quiet=False, env=None, poll=False,
            capture=None):
        """
        execute the host tool with the provided arguments (if any)

        Runs the host tool described by ``args`` until completion.

        Args:
            args (optional): the list of arguments for the tool
            cwd (optional): working directory to use
            quiet (optional): whether or not to suppress output
            env (optional): environment variables to include
            poll (optional): force polling stdin/stdout for output data
            capture (optional): list to capture output into

        Returns:
            the return code of the execution request
        """
        if not self.exists():
            return 1

        if args and not is_sequence_not_string(args):
            err('invalid argument type provided into execute (should be list): '
                + str(args))
            return 1

        final_env = None
        if self.include or self.sanitize or env:
            final_env = os.environ.copy()
            if self.sanitize:
                for key in self.sanitize:
                    final_env.pop(key, None)
            if self.include:
                final_env.update(self.include)
            if env:
                final_env.update(env)

        final_args = [self.tool]
        if args:
            final_args.extend(args)

        return _execute(final_args, cwd=cwd, env=final_env, quiet=quiet,
            critical=False, poll=poll, capture=capture)

    def exists(self):
        """
        return whether or not the host tool exists

        Returns whether or not the tool is available on the host for use.

        Returns:
            ``True``, if the tool exists; ``False`` otherwise
        """
        if self.tool in RelengTool.detected:
            return RelengTool.detected[self.tool]

        if execute([self.tool] + self.exists_args, quiet=True, critical=False):
            debug('{} tool is detected on this system', self.tool)
            RelengTool.detected[self.tool] = True
        else:
            debug('{} tool is not detected on this system', self.tool)
            RelengTool.detected[self.tool] = False

        return RelengTool.detected[self.tool]



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.cmake import CMAKE
from releng_tool.util.io import prepare_arguments
from releng_tool.util.io import prepare_definitions
from releng_tool.util.log import err
from releng_tool.util.log import verbose
from releng_tool.util.string import expand

def build(opts):
    """
    support building cmake projects

    With provided build options (``RelengBuildOptions``), the build stage will
    be processed.

    Args:
        opts: build options

    Returns:
        ``True`` if the building stage is completed; ``False`` otherwise
    """

    if not CMAKE.exists():
        err('unable to build package; cmake is not installed')
        return False

    # definitions
    cmake_defs = {
    }
    if opts.build_defs:
        cmake_defs.update(expand(opts.build_defs))

    # options
    cmake_opts = {
        # build RelWithDebInfo (when using multi-configuration projects)
        '--config': 'RelWithDebInfo',
    }
    if opts.build_opts:
        cmake_opts.update(expand(opts.build_opts))

    # argument building
    cmake_args = [
        # tell cmake to invoke build process in the output directory
        '--build',
        opts.build_output_dir,
    ]
    cmake_args.extend(prepare_definitions(cmake_defs, '-D'))
    cmake_args.extend(prepare_arguments(cmake_opts))

    # enable specific number of parallel jobs is set
    #
    # https://cmake.org/cmake/help/v3.12/manual/cmake.1.html#build-tool-mode
    if 'releng.cmake.disable_parallel_option' not in opts._quirks:
        if opts.jobsconf != 1 and opts.jobs > 1:
            cmake_args.append('--parallel')
            cmake_args.append(str(opts.jobs))
    else:
        verbose('cmake parallel jobs disabled by quirk')

    if not CMAKE.execute(cmake_args, env=expand(opts.build_env)):
        err('failed to build cmake project: {}', opts.name)
        return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2018 releng-tool

from releng_tool.util.sort import TopologicalSorter
import unittest

class TestObj:
    def __init__(self, name):
        self.children = []
        self.name = name

    def __repr__(self):
        return self.name

class TestUtilSort(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        def sorting_method(obj):
            return obj.children
        cls.sorter = TopologicalSorter(sorting_method)

    def test_utilsort_topologicalsorter_simple(self):
        """
        simple sorting tests
        """
        # a -> b -> c -> d -> e
        a = TestObj('a')
        b = TestObj('b')
        c = TestObj('c')
        d = TestObj('d')
        e = TestObj('e')
        a.children.append(b)
        b.children.append(c)
        c.children.append(d)
        d.children.append(e)

        sorted_ = self.sorter.sort(e)
        self.assertEqual(sorted_, [e])
        self.sorter.reset()

        sorted_ = self.sorter.sort(c)
        self.assertEqual(sorted_, [e, d, c])
        self.sorter.reset()

        sorted_ = self.sorter.sort(a)
        self.assertEqual(sorted_, [e, d, c, b, a])
        self.sorter.reset()

        self.sorter.sort(e)
        self.sorter.sort(b)
        self.sorter.sort(c)
        self.sorter.sort(d)
        sorted_ = self.sorter.sort(e)
        self.assertEqual(sorted_, [e, d, c, b])

    def test_utilsort_topologicalsorter_ordering(self):
        """
        ensure the order objects are provided matters (i.e. user priority)
        """
        # a, b, c
        a = TestObj('a')
        b = TestObj('b')
        c = TestObj('c')

        self.sorter.sort(a)
        self.sorter.sort(b)
        sorted_ = self.sorter.sort(c)
        self.assertEqual(sorted_, [a, b, c])
        self.sorter.reset()

        self.sorter.sort(b)
        self.sorter.sort(a)
        sorted_ = self.sorter.sort(c)
        self.assertEqual(sorted_, [b, a, c])
        self.sorter.reset()

        self.sorter.sort(c)
        self.sorter.sort(a)
        sorted_ = self.sorter.sort(b)
        self.assertEqual(sorted_, [c, a, b])
        self.sorter.reset()

        self.sorter.sort(c)
        self.sorter.sort(b)
        sorted_ = self.sorter.sort(a)
        self.assertEqual(sorted_, [c, b, a])

    def test_utilsort_topologicalsorter_multiple(self):
        """
        a bit less simple sorting test
        """
        #  /-> b
        # a
        #  \-> c -> d -> e, f, g
        a = TestObj('a')
        b = TestObj('b')
        c = TestObj('c')
        d = TestObj('d')
        e = TestObj('e')
        f = TestObj('f')
        g = TestObj('g')
        a.children.append(b)
        a.children.append(c)
        c.children.append(d)
        d.children.append(e)
        d.children.append(f)
        d.children.append(g)

        sorted_ = self.sorter.sort(a)
        self.assertEqual(sorted_, [b, e, f, g, d, c, a])

    def test_utilsort_topologicalsorter_complex(self):
        """
        complex sorting test
        """
        # a      -> f -> g
        #       /    \
        # b -> e      -> h -------------------> r
        #            /                      /
        # c -> i -> j -> x -> m, n, o, p   /
        #  /                              /
        # d ------> k -----------------> q
        #
        a = TestObj('a')
        b = TestObj('b')
        c = TestObj('c')
        d = TestObj('d')
        e = TestObj('e')
        f = TestObj('f')
        g = TestObj('g')
        h = TestObj('h')
        i = TestObj('i')
        j = TestObj('j')
        k = TestObj('k')
        x = TestObj('x')
        m = TestObj('m')
        n = TestObj('n')
        o = TestObj('o')
        p = TestObj('p')
        q = TestObj('q')
        r = TestObj('r')
        b.children.append(e)
        c.children.append(i)
        d.children.extend([i, k])
        e.children.append(f)
        f.children.extend([g, h])
        h.children.append(r)
        i.children.append(j)
        j.children.extend([h, x])
        k.children.append(q)
        x.children.extend([m, n, o, p])
        q.children.append(r)

        self.sorter.sort(a)
        self.sorter.sort(b)
        self.sorter.sort(c)
        self.sorter.sort(d)
        self.assertEqual(self.sorter.sorted,
            [a, g, r, h, f, e, b, m, n, o, p, x, j, i, c, q, k, d])

    def tearDown(self):
        self.sorter.reset()



# -*- coding: utf-8 -*-
# Copyright 2018 releng-tool

from releng_tool.util.hash import HashResult
from releng_tool.util.hash import BadFileHashLoadError
from releng_tool.util.hash import BadFormatHashLoadError
from releng_tool.util.hash import load as load_hashes
from releng_tool.util.hash import verify as verify_hashes
import os
import unittest

ASSETS_DIR = 'assets'
SAMPLES_DIR = 'hash-samples'

class TestUtilHash(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        base_dir = os.path.dirname(os.path.realpath(__file__))
        cls.samples_dir = os.path.join(base_dir, ASSETS_DIR, SAMPLES_DIR)

    def test_utilhash_loading(self):
        samples = self.samples_dir

        # generates exception if missing
        file = os.path.join(samples, 'a-file-that-does-not-exist')
        self.assertRaises(BadFileHashLoadError, load_hashes, file)

        # generates a format exception if hashes are not properly defined
        invalid_samples = [
            'invalid-hash-sample-a',
            'invalid-hash-sample-b',
            'invalid-hash-sample-c',
        ]
        for sample in invalid_samples:
            file = os.path.join(samples, sample)
            self.assertRaises(BadFormatHashLoadError, load_hashes, file)

        # test
        file = os.path.join(samples, 'valid-hash-sample-a')
        data = load_hashes(file)
        self.assertEqual(data, [
            ('type', 'hash', 'file')
        ])

        # test
        file = os.path.join(samples, 'valid-hash-sample-b')
        data = load_hashes(file)
        self.assertEqual(data, [
            ('type1', 'hash1', 'file1'),
            ('type2', 'hash2', 'file2'),
            ('type3', 'hash3', 'file3'),
            ('type4', 'hash4', 'file4'),
            ('type5', 'hash5', 'file5'),
        ])

        # test
        file = os.path.join(samples, 'valid-hash-sample-c')
        data = load_hashes(file)
        self.assertEqual(data, [
            ('md5',
                '37a460dd2f87f3c04a4a85b38d922eb1',
                'sample-asset'),
            ('sha1',
                'b713d8fc4807ff7f0b186b618ae65b6e7fa66f5a',
                'sample-asset'),
            ('sha224',
                '5d03e966de09ffc2f3c7416963c0a20631e4a950f965ffc7e6f2588d',
                'sample-asset'),
            ('sha256',
                '0f595a93a91e678d592dd7978b1d215be1adc27123509eabb6b7f02eaf1315'
                'd0a4',
                'sample-asset'),
            ('sha384',
                '60d4b57034872b8617bb1a0a0e076d0a97a6d44485c5e013e9156b8f51346d'
                '58b36ea1be51c57c23600aed1fb301e07f',
                'sample-asset'),
            ('sha512',
                '741695d2189840162a19d9a8f8c99e7adc4621f452b43d20081a5ebb23b9d7'
                '3e15c3ffb9b06c65621da269686da2d5991f5ac415cdd2cfcd5faf10d3aa47'
                'd728',
                'sample-asset'),
        ])

        # empty hash file provided an empty list
        file = os.path.join(samples, 'valid-hash-sample-d')
        data = load_hashes(file)
        self.assertEqual(data, [])

    def test_utilhash_verify(self):
        samples = self.samples_dir
        target_sample = os.path.join(samples, 'sample-asset')

        # archive-specific check
        file = os.path.join(samples, 'verify-missing-archive')
        result = verify_hashes(file, target_sample, quiet=True)
        self.assertEqual(result, HashResult.MISSING_ARCHIVE)

        file = os.path.join(samples, 'verify-success')
        result = verify_hashes(file, target_sample, quiet=True)
        self.assertEqual(result, HashResult.VERIFIED)

        # directory-specific check
        file = os.path.join(samples, 'verify-bad-format')
        result = verify_hashes(file, samples, quiet=True)
        self.assertEqual(result, HashResult.BAD_FORMAT)

        file = os.path.join(samples, 'verify-some-file-that-should-not-exist')
        result = verify_hashes(file, samples, quiet=True)
        self.assertEqual(result, HashResult.BAD_PATH)

        file = os.path.join(samples, 'verify-bad-hash')
        result = verify_hashes(file, samples, quiet=True)
        self.assertEqual(result, HashResult.MISMATCH)

        file = os.path.join(samples, 'verify-empty')
        result = verify_hashes(file, samples, quiet=True)
        self.assertEqual(result, HashResult.EMPTY)

        file = os.path.join(samples, 'verify-missing-listed')
        result = verify_hashes(file, samples, quiet=True)
        self.assertEqual(result, HashResult.MISSING_LISTED)

        file = os.path.join(samples, 'verify-success')
        result = verify_hashes(file, samples, quiet=True)
        self.assertEqual(result, HashResult.VERIFIED)

        file = os.path.join(samples, 'verify-unknown-hash-type')
        result = verify_hashes(file, samples, quiet=True)
        self.assertEqual(result, HashResult.UNSUPPORTED)

        # missing check
        result = verify_hashes('', '', quiet=True)
        self.assertEqual(result, HashResult.BAD_PATH)

        # excluded/empty
        exclude = ['sample-asset']
        file = os.path.join(samples, 'verify-success')
        result = verify_hashes(file, target_sample, exclude=exclude, quiet=True)
        self.assertEqual(result, HashResult.EMPTY)



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from releng_tool.util.io import cat
from tests import prepare_workdir
from tests import redirect_stdout
import os
import unittest


class TestUtilIoCat(unittest.TestCase):
    def test_utilio_cat_invalid_file(self):
        with redirect_stdout() as stream:
            with prepare_workdir() as test_dir:
                output = cat(test_dir)

        self.assertFalse(output)
        self.assertEqual(stream.getvalue(), '')

    def test_utilio_cat_missing_file(self):
        with redirect_stdout() as stream:
            with prepare_workdir() as test_dir:
                test_file = os.path.join(test_dir, 'missing-file')

                output = cat(test_file)

        self.assertFalse(output)
        self.assertEqual(stream.getvalue(), '')

    def test_utilio_cat_valid_contents_multiple(self):
        with redirect_stdout() as stream:
            with prepare_workdir() as test_dir:
                test_file1 = os.path.join(test_dir, 'test1')
                with open(test_file1, 'w') as f:
                    f.write('1')

                test_file2 = os.path.join(test_dir, 'test2')
                with open(test_file2, 'w') as f:
                    f.write('2')

                test_file3 = os.path.join(test_dir, 'test3')
                with open(test_file3, 'w') as f:
                    f.write('3')

                output = cat(test_file1, test_file2, test_file3)

        self.assertTrue(output)
        self.assertEqual(stream.getvalue(), '123')

    def test_utilio_cat_valid_contents_single(self):
        with redirect_stdout() as stream:
            with prepare_workdir() as test_dir:
                test_file = os.path.join(test_dir, 'test')

                with open(test_file, 'w') as f:
                    f.write('this is a test')

                output = cat(test_file)

        self.assertTrue(output)
        self.assertEqual(stream.getvalue(), 'this is a test')

    def test_utilio_cat_valid_empty(self):
        with redirect_stdout() as stream:
            with prepare_workdir() as test_dir:
                test_file = os.path.join(test_dir, 'test')

                with open(test_file, 'w'):
                    pass  # empty

                output = cat(test_file)

        self.assertTrue(output)
        self.assertEqual(stream.getvalue(), '')



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

# sample extension will should fail to load since the the required
# version number is newer than the current version
def releng_setup(app):
    app.require_version('999999')



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from tests import prepare_testenv
from tests.support import fetch_unittest_assets_dir
import os
import unittest


class TestEngineSubdir(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.template_base = 'subdir'

        sample_files = fetch_unittest_assets_dir('sample-files')
        archive = os.path.join(sample_files, 'sample-files.tar')
        cls.site = 'TEST_SITE=r"file:///{}"'.format(archive)

    def test_engine_run_subdir_valid_bootstrap(self):
        template = os.path.join(self.template_base, 'subdir-valid-bootstrap')
        with prepare_testenv(template=template) as engine:
            root_dir = engine.opts.root_dir
            pkg_script = os.path.join(root_dir, 'package', 'test', 'test')

            with open(pkg_script, 'a') as f:
                f.write(self.site + '\n')

            rv = engine.run()
            self.assertTrue(rv)

            flag = os.path.join(engine.opts.target_dir, 'invoked-bootstrap')
            self.assertTrue(os.path.exists(flag))

    def test_engine_run_subdir_valid_configure(self):
        template = os.path.join(self.template_base, 'subdir-valid-configure')
        with prepare_testenv(template=template) as engine:
            root_dir = engine.opts.root_dir
            pkg_script = os.path.join(root_dir, 'package', 'test', 'test')

            with open(pkg_script, 'a') as f:
                f.write(self.site + '\n')

            rv = engine.run()
            self.assertTrue(rv)

            flag = os.path.join(engine.opts.target_dir, 'invoked-configure')
            self.assertTrue(os.path.exists(flag))

    def test_engine_run_subdir_valid_build(self):
        template = os.path.join(self.template_base, 'subdir-valid-build')
        with prepare_testenv(template=template) as engine:
            root_dir = engine.opts.root_dir
            pkg_script = os.path.join(root_dir, 'package', 'test', 'test')

            with open(pkg_script, 'a') as f:
                f.write(self.site + '\n')

            rv = engine.run()
            self.assertTrue(rv)

            flag = os.path.join(engine.opts.target_dir, 'invoked-build')
            self.assertTrue(os.path.exists(flag))

    def test_engine_run_subdir_valid_install(self):
        template = os.path.join(self.template_base, 'subdir-valid-install')
        with prepare_testenv(template=template) as engine:
            root_dir = engine.opts.root_dir
            pkg_script = os.path.join(root_dir, 'package', 'test', 'test')

            with open(pkg_script, 'a') as f:
                f.write(self.site + '\n')

            rv = engine.run()
            self.assertTrue(rv)

            flag = os.path.join(engine.opts.target_dir, 'invoked-install')
            self.assertTrue(os.path.exists(flag))

    def test_engine_run_subdir_valid_post(self):
        template = os.path.join(self.template_base, 'subdir-valid-post')
        with prepare_testenv(template=template) as engine:
            root_dir = engine.opts.root_dir
            pkg_script = os.path.join(root_dir, 'package', 'test', 'test')

            with open(pkg_script, 'a') as f:
                f.write(self.site + '\n')

            rv = engine.run()
            self.assertTrue(rv)

            flag = os.path.join(engine.opts.target_dir, 'invoked-post')
            self.assertTrue(os.path.exists(flag))



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from tests import prepare_testenv
from tests import prepare_workdir
import os
import unittest

class TestEngineRunEnviron(unittest.TestCase):

    def run(self, result=None):
        old_env = dict(os.environ)
        try:
            super(TestEngineRunEnviron, self).run(result)
        finally:
            os.environ.clear()
            os.environ.update(old_env)

    def test_engine_run_environ_assets_dir(self):
        with prepare_workdir() as assets_dir:
            os.environ['RELENG_ASSETS_DIR'] = assets_dir

            with prepare_testenv() as engine:
                expected_cache_dir = os.path.join(assets_dir, 'cache')
                expected_dl_dir = os.path.join(assets_dir, 'dl')
                self.assertEqual(engine.opts.cache_dir, expected_cache_dir)
                self.assertEqual(engine.opts.dl_dir, expected_dl_dir)

    def test_engine_run_environ_cache_dir(self):
        with prepare_workdir() as cache_dir:
            os.environ['RELENG_CACHE_DIR'] = cache_dir

            with prepare_testenv() as engine:
                self.assertEqual(engine.opts.cache_dir, cache_dir)

    def test_engine_run_environ_dl_dir(self):
        with prepare_workdir() as dl_dir:
            os.environ['RELENG_DL_DIR'] = dl_dir

            with prepare_testenv() as engine:
                self.assertEqual(engine.opts.dl_dir, dl_dir)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

import ctypes
import os
import re
import subprocess

try:
    import ctypes.wintypes

    try:
        import winreg as reg
    except ImportError:
        import _winreg as reg
except Exception:
    pass

def enable_ansi():
    """
    enable ansi in a windows command window

    Adjusts a Windows command window to support parsing control characters
    pushed to output handle. The primary goal is to support color control
    characters generated by releng-tool's logging events.
    """

    kern32 = ctypes.windll.kernel32

    # fetch the current console mode
    mode = ctypes.wintypes.DWORD()
    handle = kern32.GetStdHandle(subprocess.STD_OUTPUT_HANDLE)
    kern32.GetConsoleMode(handle, ctypes.byref(mode))

    # adjust the console mode to parse output in the console
    # https://docs.microsoft.com/en-us/windows/console/setconsolemode
    ENABLE_VIRTUAL_TERMINAL_PROCESSING = 7
    mode = mode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING
    kern32.SetConsoleMode(handle, mode)

def find_win32_python_interpreter(target):
    """
    find a python interpreter on a windows system

    Attempts to find the full path of a Python interpreter on a Windows system
    by checking the running context's registry. This call will cycle through
    various registry hives and match an installed interpreter based off the
    provided target string. For example, providing a ``target`` value of
    ``python3.9`` (or ``3.9``) will attempt to find the first available Python
    3.9 interpreter on the system. When providing a non-specific Python version
    (e.g. ``python3``), the most recent version of the interpreter type will be
    used). User-installed interpreters are promoted over shared interpreters.

    Args:
        target: the target interpreter

    Returns:
        the interpreter path; ``None`` if no interpreter could be found
    """

    if not target:
        return None

    PYTHON_PREFIX = 'python'
    PYTHON_CORE = r'Software\Python\PythonCore'
    INSTALL_PATH_KEY = r'\InstallPath'
    PYTHON_EXEC = r'python.exe'

    # remove any python prefix ('python3.8' -> '3.8')
    if target.startswith(PYTHON_PREFIX):
        target = target[len(PYTHON_PREFIX):]

    # keep track of a dictionary of available interpreters when trying to pick
    # a provided major version (e.g. "python3"), when cycling through all the
    # available registry hives
    found_interpreters = {}

    # see also https://www.python.org/dev/peps/pep-0514/
    for hive, key, flags in [
        (reg.HKEY_CURRENT_USER, PYTHON_CORE, 0),
        (reg.HKEY_LOCAL_MACHINE, PYTHON_CORE, reg.KEY_WOW64_64KEY),
        (reg.HKEY_LOCAL_MACHINE, PYTHON_CORE, reg.KEY_WOW64_32KEY),
    ]:
        try:
            # search for official python releases
            with reg.OpenKeyEx(hive, key, 0, reg.KEY_READ | flags) as root:
                # compile a list of known version
                i = 0
                installed_vers = []
                raw_interpreters = {}
                while True:
                    try:
                        raw_interpreter = reg.EnumKey(root, i)

                        # remove any architecture that may be registered for
                        # the detected interpreter; e.g. "python37-32"
                        interpreter = raw_interpreter.split('-', 1)[0]
                        raw_interpreters[interpreter] = raw_interpreter

                        installed_vers.append(interpreter)
                    except OSError:
                        break
                    i += 1
                if not installed_vers:
                    continue

                # sort by newest interpreter
                _sort_interpreters(installed_vers)

                # find a version which matches are desired target
                target_ver = None
                if target in installed_vers:
                    target_ver = target
                # first available version?
                elif target == '':
                    target_ver = installed_vers[0]
                # if we do have an explicit match and only a major version is
                # provided, attempt to find a first available (most recent)
                # interpreter available on the system
                elif '.' not in target:
                    target_major = target + '.'

                    if any(v.startswith(target_major) for v in installed_vers):
                        for installed_ver in installed_vers:
                            if installed_ver.startswith(target_major):
                                target_ver = installed_ver
                                break

                # skip root if target version cannot be found
                if not target_ver:
                    continue

                # skip version if already found in another hive
                if target_ver in found_interpreters:
                    continue

                # load key to find the interpreter's path
                base_key = raw_interpreters[target_ver]
                try:
                    with reg.OpenKey(root, base_key + INSTALL_PATH_KEY) as k:
                        try:
                            install_path = reg.QueryValue(k, '')
                            full_path = os.path.join(install_path, PYTHON_EXEC)

                            # rule out the registry item if the path/executable
                            # does not exists
                            if not os.path.exists(full_path):
                                continue

                            # if matched to a target, stop
                            if target and target_ver == target:
                                return full_path

                            found_interpreters[target_ver] = full_path
                            continue
                        except ImportError:
                            pass
                except OSError:
                    pass
        except OSError:
            pass

    # sort by newest interpreter
    if found_interpreters:
        installed_vers = list(found_interpreters.keys())
        _sort_interpreters(installed_vers)
        return found_interpreters[installed_vers[0]]

    return None

def _sort_interpreters(o):
    """
    in-place sort a list of interpreters

    A helper method (to be used with ``find_interpreter``) to help sort a list
    of interpreters by comparing and prioritizing detected version strings.
    Newer interpreters will be sorted before older interpreters.

    Args:
        o: the list to sort
    """

    o.sort(
        key=lambda f: [int(x) for x in re.sub(r'[^\d\.]', '', f).split('.')],
        reverse=True)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

def nullish_coalescing(value, default):
    """
    nullish coalescing utility call

    Provides a return of the provided value unless the value is a ``None``,
    which instead the provided default value is returned instead.

    Args:
        value: the value
        default: the default value

    Returns:
        the provided value if not ``None``; otherwise, the default value
    """
    if value is None:
        return default

    return value



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from __future__ import absolute_import
from io import open
from releng_tool.util.enum import Enum
from releng_tool.util.log import debug
from releng_tool.util.log import err
from releng_tool.util.log import warn
import hashlib
import os

#: size of blocks read when calculating the hash for a file
HASH_READ_BLOCKSIZE = 1048576

class HashResult(Enum):
    """
    result of a hash verification check

    Attributes:
        BAD_FORMAT: a bad format was detected in hash file
        BAD_PATH: a path is missing or cannot be read
        EMPTY: the hash file is empty
        MISMATCH: detected a mismatch between provided and calculated hash
        MISSING_ARCHIVE: missing hash for archive
        MISSING_LISTED: a listed file is missing in sources
        UNSUPPORTED: an unsupported hash type was provided
        VERIFIED: hash verification has completed
    """
    BAD_FORMAT = 'bad_format'
    BAD_PATH = 'bad_path'
    EMPTY = 'empty'
    MISMATCH = 'mismatch'
    MISSING_ARCHIVE = 'missing_archive'
    MISSING_LISTED = 'missing_listed'
    UNSUPPORTED = 'unsupported'
    VERIFIED = 'verified'

class BadFileHashLoadError(Exception):
    """
    raised when a file could not be loaded for hash information
    """

class BadFormatHashLoadError(Exception):
    """
    raised when invalid format/contents is detected in a hash file
    """

def load(hash_file):
    """
    load hash entries from a provided file

    The UTF-8 encoded file ``hash_file`` will be loaded for defined hash
    entries. A hash entry is a 3-tuple defining the type of hash algorithm used,
    the hash value expected and the asset associated with the hash. A tuple
    entry is defined on a single line with each entry separated by whitespace
    characters.

    A file with no entries defined will return an empty list.

    Comments are permitted in the file. Lines leading with a ``#`` character or
    inlined leading ``#`` character after a whitespace character will be
    ignored.

    Args:
        hash_file: the file

    Returns:
        a list of hash entry tuples (if any exist)

    Raises:
        BadFileHashLoadError: the hash file cannot be found or loaded
        BadFormatHashLoadError: a format issue has been detected
    """
    try:
        with open(hash_file, mode='r', encoding='utf_8') as f:
            data = f.readlines()
    except IOError:
        raise BadFileHashLoadError()

    # strip and split into chunks
    data = [x.split() for x in data if x.strip()]

    for idx, entry in enumerate(data):
        # remove comments
        for idx2, parts in enumerate(entry):
            if parts.startswith('#'):
                del entry[idx2:]
                break
        if not entry:
            continue

        entry_len = len(entry)
        if entry_len > 3:
            raise BadFormatHashLoadError(
                'too many values for entry {}'.format(idx + 1))
        if entry_len < 3:
            raise BadFormatHashLoadError(
                'too few values for entry {}'.format(idx + 1))

    # compile a list of tuples with no empty entries
    return [tuple(x) for x in data if x]

def verify(hash_file, path, exclude=None, relaxed=False, quiet=False):
    """
    verify a file or directory with the hashes defined in the provided hash file

    Performs a hash verification check for one or more files at the provided
    path. The provided ``hash_file`` will contain a series of expected hash
    entries for one or more files. If the provided ``path`` is a file, the
    file's hash will be compared to a matching entry in the hash file. If the
    provided ``path` is a directory, each entry defined in the ``hash_file``
    will be checked and verified. The ``hash_file`` is a UTF-8 encoded file
    containing 3-tuple entries defining the type of hash algorithm used, the
    hash value expected and the asset associated with the hash (see
    ``load_hashes``). On the successful verification of all files, this method
    will return ``HashResult.VERIFIED``. Other warning/error states for
    verification will be return with a respective ``HashResult`` value.

    Args:
        hash_file: the file containing hash information
        path: the file or directory to verify
        exclude: assets to exclude from the verification check
        relaxed: relax logging to only warn on detected missing/mismatched
        quiet: disablement of error messages to standard out

    Returns:
        the hash result (``HashResult``)
    """
    is_file = False
    if os.path.isdir(path):
        pass
    elif os.path.isfile(path):
        is_file = True
    else:
        return HashResult.BAD_PATH

    # load hashes
    try:
        hash_info = load(hash_file)
    except BadFileHashLoadError:
        return HashResult.BAD_PATH
    except BadFormatHashLoadError as e:
        if not quiet:
            err('''\
hash file is not properly formatted

The hash file provided is incorrectly formatted. The hash file expects lines
with the hash type, hash and target file provided. For example:

    sha1 f572d396fae9206628714fb2ce00f72e94f2258f my-file

Please correct the following hash file:

    Hash File: {}
      Details: {}''', hash_file, e)

        return HashResult.BAD_FORMAT

    # no hash information
    if not hash_info:
        return HashResult.EMPTY

    # if this is a target file, filter out other hash entries
    if is_file:
        target = os.path.basename(path)
        path = os.path.abspath(os.path.join(path, os.pardir))
        hash_info = [x for x in hash_info if x[2] == target]
        if not hash_info:
            return HashResult.MISSING_ARCHIVE

    # filter out excluded assets (if any)
    if exclude:
        hash_info = [x for x in hash_info if x[2] not in exclude]
        if not hash_info:
            return HashResult.EMPTY

    hash_catalog = {}
    for type_, hash_, asset in hash_info:
        types = hash_catalog.setdefault(asset, {})
        types.setdefault(type_, []).append(hash_.lower())

    for asset, type_hashes in hash_catalog.items():
        hashers = {}
        for hash_type in type_hashes.keys():
            hashers[hash_type] = _get_hasher(hash_type)
            if not hashers[hash_type]:
                if not quiet:
                    err('''\
unsupported hash type

The hash file defines a hash type not supported by the releng-tool. Officially
supported hash types are FIPS-180 algorithms (sha1, sha224, sha256, sha384 and
sha512) as well as (but not recommended) RSA'S MD5 algorithm. Other algorithms,
while unofficially supported, can be used if provided by the system's OpenSSL
library.

     Hash File: {}
 Provided Type: {}''', hash_file, type)

                return HashResult.UNSUPPORTED

        target_file = os.path.join(path, asset)
        try:
            with open(target_file, 'rb') as f:
                buf = f.read(HASH_READ_BLOCKSIZE)
                while buf:
                    for hasher in hashers.values():
                        hasher.update(buf)
                    buf = f.read(HASH_READ_BLOCKSIZE)
        except IOError:
            if not quiet:
                if relaxed:
                    warn('missing expected file for verification: ' + asset)
                else:
                    err('''\
missing expected file for verification

A defined hash entry cannot be verified since the target file does not exist.
Ensure the hash file correctly names an expected file.

    Hash File: {}
         File: {}''', hash_file, asset)

            return HashResult.MISSING_LISTED

        for hash_type, hasher in hashers.items():
            digest = hasher.hexdigest()
            debug('calculated-hash: {} {}:{}', asset, hash_type, digest)
            hashes = type_hashes[hash_type]
            if digest not in hashes:
                if not quiet:
                    if relaxed:
                        warn('hash mismatch detected: ' + asset)
                    else:
                        provided = ''
                        for hash_ in hashes:
                            provided += '\n     Provided: {}'.format(hash_)

                        err('''\
hash mismatch detected

    Hash File: {}
         File: {}
     Detected: {}{}''', hash_file, asset, digest, provided)

                return HashResult.MISMATCH

    return HashResult.VERIFIED

def _get_hasher(type_):
    """
    obtain a hasher instance from the provided type

    Finds and returns an instance which is capable of performing hashing
    operations of an algorithm defined by the ``type`` provided. In the event
    that a hasher cannot be found, this method will return ``None``.

    Args:
        type_: the type of hash

    Returns:
        the hasher instance; ``None`` if the hash type is not supported
    """
    type_ = type_.lower()
    func = getattr(hashlib, type_, None)
    if func:
        return func()
    return None



# -*- coding: utf-8 -*-
# Copyright 2018 releng-tool

class TopologicalSorter:
    """
    utility used to generate a topological sorted list

    The goal of this utility is to generate a topological sorted list (using
    depth-first search) based on one or more provided nodes in a prepared graph.
    When ``sort`` is invoked, the sorter will check the graph to leaf nodes and
    visit nodes until a sorted list can be returned. Leaf nodes are determined
    by the provided sorting function ``sort_func`` which will expect a returned
    collection of leaf nodes for an object in  a graph. This sorting method will
    return an order prioritizing leaf nodes instead of parent nodes. The sorter
    can handle being passed multiple nodes in a graph at any time; however, only
    if the graph's structure does not change (i.e. graph edges are not changed).

    Args:
        sort_func: a function to return leaf nodes for a node being
            visited during the sorting process

    Attributes:
        sorted: list of currently sorted objects
    """
    def __init__(self, sort_func):
        assert sort_func, 'no sort function provided'
        self.sorted = []
        self._state = {}
        self._sort_func = sort_func

    def sort(self, obj):
        """
        generate/update a topological sorted list from a object in a graph

        From a provided graph object ``obj``, a topological sorted (depth-first
        search) list will be updated and returned. This sort call can be invoked
        multiple times.

        Args:
            obj: vertex in a graph to sort from

        Returns:
            sorted list of objects; ``None`` if a cyclic graph has been
                detected

            The returned list of objects should not be directly modified until
            all sorting calls (if multiple sort operations a desired) are
            completed.
        """
        if self._visit(obj):
            return self.sorted
        return None

    def reset(self):
        """
        reset the state of the sorter

        Resets tracked state information contained in the sorter and clears the
        known sorted list of vertices.
        """
        self.sorted = []
        self._state = {}

    def _visit(self, obj):
        """
        depth-first search topological sort visit call

        From a provided graph object ``obj``, a topological sorted (depth-first
        search) list will be updated and returned. This sort call can be invoked
        multiple times.

        Args:
            obj: vertex in a graph to sort from

        Returns:
            ``True``, if the sorting was successful; ``False``, if a cyclic
                graph has been detected
        """
        if obj not in self._state:
            self._state[obj] = ''
        if self._state[obj] == 'P':
            return True
        if self._state[obj] == 'T':
            return False
        self._state[obj] = 'T'
        for child in self._sort_func(obj):
            if not self._visit(child):
                return False
        self._state[obj] = 'P'
        self.sorted.append(obj)
        return True



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from contextlib import contextmanager
from tests.support import fetch_unittest_assets_dir
from threading import Thread
import os
import ssl
import time

try:
    import http.server as http_server
except ImportError:
    import BaseHTTPServer as http_server


class MockServerRequestHandler(http_server.BaseHTTPRequestHandler):
    """
    mock server request handler

    Provides a handler for managing HTTP requests and responses for unit
    testing.
    """

    def do_GET(self):
        self._track_request('GET')
        self._process_rsp()

    def do_HEAD(self):
        self._track_request('HEAD')
        self._process_rsp()

    def do_POST(self):
        self._track_request('POST')
        self._process_rsp()

    def do_PUT(self):
        self._track_request('PUT')
        self._process_rsp()

    def _process_rsp(self):
        try:
            code, data = self.server.rsp.pop(0)
        except IndexError:
            code = 500
            data = None

        self.send_response(code)
        self.end_headers()
        if data:
            self.wfile.write(data)

    def _track_request(self, action):
        requests = self.server.req.setdefault(action, [])
        requests.append((self.path, dict(self.headers)))


def build_httpd(secure=None):
    """
    build an http daemon

    Builds an HTTP server instance on a random local port which can be used
    to verify HTTP/HTTPS request handling for various releng-tool features.
    By default, an HTTP is spawned with `secure` set to `False`. When building
    a "secure" HTTP server, key/cert. files are used from the assets folder.

    Args:
        secure (optional): whether or not a secure http server should be made

    Returns:
        the http server
    """

    LOCAL_RANDOM_PORT = ('127.0.0.1', 0)
    httpd = http_server.HTTPServer(LOCAL_RANDOM_PORT, MockServerRequestHandler)
    httpd.req = {}
    httpd.rsp = []

    if secure:
        keyfile, certfile = fetch_cert_files()

        context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        context.load_cert_chain(certfile, keyfile=keyfile)

        httpd.socket = context.wrap_socket(httpd.socket, server_side=True)

    return httpd


def fetch_cert_files():
    """
    find certificate files for unit testing

    Returns a tuple of a key file and certificate file used to create a
    unit testing "secure" HTTP server. Dummy pem files are included in the
    testing asserts folder, which this call can return the paths to these
    files.

    Returns:
        2-tuple of a key file and certificate file
    """

    httpd_assets = fetch_unittest_assets_dir('httpd')
    keyfile = os.path.join(httpd_assets, 'test-notprivate-key-pem')
    certfile = os.path.join(httpd_assets, 'test-cert-pem')
    return keyfile, certfile


@contextmanager
def httpd_context(secure=None):
    """
    create an http daemon context

    Builds a context-enabled HTTP server instance on a random local port which
    can be used to verify HTTP/HTTPS request handling for various releng-tool
    features. The build HTTP server is served by an internally managed thread.
    By default, an HTTP is spawned with `secure` set to `False`. When building
    a "secure" HTTP server, key/cert. files are used from the assets folder.

    Args:
        secure (optional): whether or not a secure http server should be made

    Yields:
        the http server
    """

    httpd = None
    httpd_thread = None

    try:
        httpd = build_httpd(secure=secure)

        # start accepting requests
        def serve_forever(httpd):
            httpd.serve_forever()

        httpd_thread = Thread(target=serve_forever, args=(httpd,))
        httpd_thread.start()

        # yield context for a moment to help threads to ready up
        time.sleep(0)

        yield httpd

    finally:
        if httpd_thread:
            httpd.shutdown()
            httpd_thread.join()
        elif httpd:
            httpd.socket.close()



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool

#: executable used to run cmake commands
CMAKE_COMMAND = 'cmake'

#: cmake host tool helper
CMAKE = RelengTool(CMAKE_COMMAND)



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from tests import prepare_testenv
import os
import unittest

class TestEngineRunFileFlags(unittest.TestCase):
    def test_engine_run_file_flag_devmode(self):
        config = {
            'development': True,
        }

        with prepare_testenv(config=config, template='minimal') as engine:
            engine.run()
            self.assertTrue(os.path.exists(engine.opts.ff_devmode))

    def test_engine_run_file_flag_local_srcs(self):
        config = {
            'local_sources': True,
        }

        with prepare_testenv(config=config, template='minimal') as engine:
            engine.run()
            self.assertTrue(os.path.exists(engine.opts.ff_local_srcs))



# -*- coding: utf-8 -*-
# Copyright 2019-2021 releng-tool

from releng_tool.util.io import interim_working_dir
from releng_tool.util.io import opt_file
from releng_tool.util.io import run_script
from releng_tool.util.log import verbose
import os
import sys

#: filename of the script to execute the post-processing operation (if any)
POST_SCRIPT = 'post'

def stage(engine, pkg, script_env):
    """
    handles the post-processing stage for a package

    With a provided engine and package instance, the post-processing stage will
    be processed. This stage is typically not advertised and is for advanced
    cases where a developer wishes to manipulate their build environment after
    package has completed each of its phases.

    Args:
        engine: the engine
        pkg: the package being built
        script_env: script environment information

    Returns:
        ``True`` if the post-processing stage is completed; ``False`` otherwise
    """

    verbose('post-processing {}...', pkg.name)
    sys.stdout.flush()

    post_script_filename = '{}-{}'.format(pkg.name, POST_SCRIPT)
    post_script = os.path.join(pkg.def_dir, post_script_filename)
    post_script, post_script_exists = opt_file(post_script)
    if not post_script_exists:
        return True

    if pkg.build_subdir:
        build_dir = pkg.build_subdir
    else:
        build_dir = pkg.build_dir

    with interim_working_dir(build_dir):
        if not run_script(post_script, script_env, subject='post-processing'):
            return False

    verbose('post-processing script executed: ' + post_script)
    return True



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool

#: executable used to run mercurial commands
HG_COMMAND = 'hg'

#: list of environment keys to filter from a environment dictionary
HG_SANITIZE_ENV_KEYS = [
    'HGENCODING',
    'HGENCODINGMODE',
    'HGENCODINGAMBIGUOUS',
    'HGRCPATH',
    'HGPLAIN',
    'HGPLAINEXCEPT',
]

#: dictionary of environment entries append to the environment dictionary
HG_EXTEND_ENV = {
    # hg is most likely a python script; ensure output is unbuffered
    'PYTHONUNBUFFERED': '1',
}

#: mercurial host tool helper
HG = RelengTool(HG_COMMAND,
    env_sanitize=HG_SANITIZE_ENV_KEYS, env_include=HG_EXTEND_ENV)



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from releng_tool.packages import pkg_cache_key
from releng_tool.tool.git import GIT
from releng_tool.util.log import debug
from releng_tool.util.log import err
from releng_tool.util.log import log
from releng_tool.util.log import warn
import os

def extract(opts):
    """
    support extraction (checkout) of a git cache into a build directory

    With provided extraction options (``RelengExtractOptions``), the extraction
    stage will be processed. A Git extraction process will populate a working
    tree based off the cached Git tree acquired from the fetch stage.

    Args:
        opts: the extraction options

    Returns:
        ``True`` if the extraction stage is completed; ``False`` otherwise
    """

    assert opts
    cache_dir = opts.cache_dir
    revision = opts.revision
    work_dir = opts.work_dir

    if not GIT.exists():
        err('unable to extract package; git is not installed')
        return None

    # extract the package
    if not _workdir_extract(cache_dir, work_dir, revision):
        return False

    # extract submodules (if configured to do so)
    if opts._git_submodules:
        if not _process_submodules(opts, work_dir):
            return False

    return True

def _process_submodules(opts, work_dir):
    """
    process submodules for an extracted repository

    After extracting a repository to a working tree, this call can be used to
    extract any tracked submodules configured on the repository. The
    ``.gitmodules`` file is parsed for submodules and caches will be populated
    for each submodule. This call is recursive.

    Args:
        opts: the extraction options
        work_dir: the working directory to look for submodules

    Returns:
        ``True`` if submodules have been processed; ``False`` otherwise
    """

    git_modules_file = os.path.join(work_dir, '.gitmodules')
    if not os.path.exists(git_modules_file):
        return True

    debug('parsing git submodules file: {}', git_modules_file)
    cfg = GIT.parse_cfg_file(git_modules_file)
    if not cfg:
        err('failed to parse git submodule')
        return False

    for sec_name in cfg.sections():
        if not sec_name.startswith('submodule'):
            continue

        if not cfg.has_option(sec_name, 'path') or \
                not cfg.has_option(sec_name, 'url'):
            debug('submodule section missing path/url')
            continue

        submodule_path = cfg.get(sec_name, 'path')
        submodule_revision = None
        if cfg.has_option(sec_name, 'branch'):
            submodule_revision = cfg.get(sec_name, 'branch')
        submodule_url = cfg.get(sec_name, 'url')
        log('extracting submodule ({}): {}', opts.name, submodule_path)
        debug('submodule revision: {}',
            submodule_revision if submodule_revision else '(none)')

        ckey = pkg_cache_key(submodule_url)
        root_cache_dir = os.path.abspath(
            os.path.join(opts.cache_dir, os.pardir))
        sm_cache_dir = os.path.join(root_cache_dir, ckey)

        postfix_path = os.path.split(submodule_path)
        sm_work_dir = os.path.join(work_dir, *postfix_path)

        if not _workdir_extract(sm_cache_dir, sm_work_dir, submodule_revision):
            return False

        # process nested submodules
        if not _process_submodules(opts, sm_work_dir):
            return False

    return True

def _workdir_extract(cache_dir, work_dir, revision):
    """
    extract a provided revision from a cache (bare) repository to a work tree

    Using a provided bare repository (``cache_dir``) and a working tree
    (``work_dir``), extract the contents of the repository using the providing
    ``revision`` value. This call will force the working directory to match the
    target revision. In the case where the work tree is diverged, the contents
    will be replaced with the origin's revision.

    Args:
        cache_dir: the cache repository
        work_dir: the working directory
        revision: the revision

    Returns:
        ``True`` if the extraction has succeeded; ``False`` otherwise
    """

    git_dir = '--git-dir=' + cache_dir
    work_tree = '--work-tree=' + work_dir

    # if a revision is not provided, extract the HEAD from the cache
    if not revision:
        revision = GIT.extract_submodule_revision(cache_dir)
        if not revision:
            return False

    log('checking out target revision into work tree')
    if not GIT.execute([git_dir, work_tree, '-c', 'advice.detachedHead=false',
                'checkout', '--force', revision],
            cwd=work_dir):
        err('unable to checkout revision')
        return False

    log('ensure target revision is up-to-date in work tree')
    origin_revision = 'origin/{}'.format(revision)
    output = []
    if GIT.execute([git_dir, 'rev-parse', '--quiet', '--verify',
            origin_revision], quiet=True, capture=output):
        remote_revision = ''.join(output)

        output = []
        GIT.execute([git_dir, 'rev-parse', '--quiet', '--verify', 'HEAD'],
            quiet=True, capture=output)
        local_revision = ''.join(output)

        debug('remote revision: {}', remote_revision)
        debug('local revision: {}', local_revision)

        if local_revision != remote_revision:
            warn('diverged revision detected; attempting to correct...')
            if not GIT.execute(
                    [
                        git_dir,
                        work_tree,
                        'reset',
                        '--hard',
                        origin_revision,
                    ], cwd=work_dir):
                err('unable to checkout revision')
                return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool
import sys

#: executable used to run bzr commands
BZR_COMMAND = 'bzr'

#: list of environment keys to filter from a environment dictionary
BZR_SANITIZE_ENV_KEYS = [
    # prevent debugger from loading
    'BZR_PDB',
    # misc
    'BZR_REMOTE_PATH',
    'BZR_TEXTUI_INPUT',
]

#: dictionary of environment entries append to the environment dictionary
BZR_EXTEND_ENV = {
    # always use a progress bar
    'BZR_PROGRESS_BAR': 'text',
}

# always suppress file logging
if sys.platform != 'win32':
    BZR_EXTEND_ENV['BZR_LOG'] = '/dev/null'
else:
    BZR_EXTEND_ENV['BZR_LOG'] = 'NUL'

#: bzr host tool helper
BZR = RelengTool(BZR_COMMAND,
    env_sanitize=BZR_SANITIZE_ENV_KEYS, env_include=BZR_EXTEND_ENV)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.api import RelengInstallOptions
from releng_tool.defs import PackageInstallType
from releng_tool.defs import PackageType
from releng_tool.engine.autotools.install import install as install_autotools
from releng_tool.engine.cmake.install import install as install_cmake
from releng_tool.engine.python.install import install as install_python
from releng_tool.engine.script.install import install as install_script
from releng_tool.util import nullish_coalescing as NC
from releng_tool.util.api import replicate_package_attribs
from releng_tool.util.io import interim_working_dir
from releng_tool.util.log import err
from releng_tool.util.log import note
import sys

def stage(engine, pkg, script_env):
    """
    handles the installation stage for a package

    With a provided engine and package instance, the installation stage will be
    processed.

    Args:
        engine: the engine
        pkg: the package being built
        script_env: script environment information

    Returns:
        ``True`` if the installation stage is completed; ``False`` otherwise
    """

    note('installing {}...', pkg.name)
    sys.stdout.flush()

    if pkg.build_subdir:
        build_dir = pkg.build_subdir
    else:
        build_dir = pkg.build_dir

    pkg_install_type = NC(pkg.install_type, PackageInstallType.TARGET)

    if pkg_install_type == PackageInstallType.HOST:
        dest_dirs = [engine.opts.host_dir]
    elif pkg_install_type == PackageInstallType.IMAGES:
        dest_dirs = [engine.opts.images_dir]
    elif pkg_install_type == PackageInstallType.STAGING:
        dest_dirs = [engine.opts.staging_dir]
    elif pkg_install_type == PackageInstallType.STAGING_AND_TARGET:
        dest_dirs = [engine.opts.staging_dir, engine.opts.target_dir]
    else:
        # default to target directory
        dest_dirs = [engine.opts.target_dir]

    install_opts = RelengInstallOptions()
    replicate_package_attribs(install_opts, pkg)
    install_opts.build_dir = build_dir
    install_opts.build_output_dir = pkg.build_output_dir
    install_opts.cache_file = pkg.cache_file
    install_opts.def_dir = pkg.def_dir
    install_opts.dest_dirs = dest_dirs
    install_opts.env = script_env
    install_opts.ext = pkg.ext_modifiers
    install_opts.host_dir = engine.opts.host_dir
    install_opts.images_dir = engine.opts.images_dir
    install_opts.install_defs = pkg.install_defs
    install_opts.install_env = pkg.install_env
    install_opts.install_opts = pkg.install_opts
    install_opts.install_type = pkg_install_type
    install_opts.name = pkg.name
    install_opts.prefix = NC(pkg.prefix, engine.opts.sysroot_prefix)
    install_opts.staging_dir = engine.opts.staging_dir
    install_opts.symbols_dir = engine.opts.symbols_dir
    install_opts.target_dir = engine.opts.target_dir
    install_opts.version = pkg.version
    install_opts._quirks = engine.opts.quirks

    installer = None
    if pkg.type in engine.registry.package_types:
        def _(opts):
            return engine.registry.package_types[pkg.type].install(
                pkg.type, opts)
        installer = _
    elif pkg.type == PackageType.AUTOTOOLS:
        installer = install_autotools
    elif pkg.type == PackageType.CMAKE:
        installer = install_cmake
    elif pkg.type == PackageType.PYTHON:
        installer = install_python
    elif pkg.type == PackageType.SCRIPT:
        installer = install_script

    if not installer:
        err('installer type is not implemented: {}', pkg.type)
        return False

    with interim_working_dir(build_dir):
        installed = installer(install_opts)
        if not installed:
            return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from releng_tool.packages.pipeline import RelengPackagePipeline
from tests import prepare_testenv
from tests.support.env_test import EnvironmentTestCase
import json
import os


class TestPkgPipelinePackageEnvironment(EnvironmentTestCase):
    def test_pkg_pipeline_pkgenv_empty(self):
        pkg_name = 'test-empty'
        with prepare_testenv(template='pkg-env') as engine:
            pkgs = engine.pkgman.load([pkg_name])

            pipeline = RelengPackagePipeline(engine, engine.opts, {})
            for pkg in pkgs:
                processed = pipeline.process(pkg)
                self.assertTrue(processed)

            # verify package-specific environment variables were set
            self.assertTrue('TEST_EMPTY_BUILD_OUTPUT_DIR' in os.environ)
            test_outdir = os.environ['TEST_EMPTY_BUILD_OUTPUT_DIR']
            results = os.path.join(test_outdir, 'invoke-env.json')
            self.assertTrue(os.path.exists(results))

            with open(results, 'r') as f:
                data = json.load(f)

            # verify that provided environment keys which may be empty if not
            # set are indeed empty
            self.assertEqual(data['PKG_SITE'], '')
            self.assertEqual(data['PKG_VERSION'], '')

    def test_pkg_pipeline_pkgenv_generic(self):
        pkg_name = 'test-generic'
        with prepare_testenv(template='pkg-env') as engine:
            pkgs = engine.pkgman.load([pkg_name])

            pipeline = RelengPackagePipeline(engine, engine.opts, {})
            for pkg in pkgs:
                processed = pipeline.process(pkg)
                self.assertTrue(processed)

            # verify package-specific environment variables were set
            self.assertTrue('TEST_GENERIC_BUILD_OUTPUT_DIR' in os.environ)
            test_outdir = os.environ['TEST_GENERIC_BUILD_OUTPUT_DIR']
            results = os.path.join(test_outdir, 'invoke-env.json')
            self.assertTrue(os.path.exists(results))

            with open(results, 'r') as f:
                data = json.load(f)

            # verify that all expected environment variables are set
            expected_pkg_env_keys = [
                'PKG_BUILD_BASE_DIR',
                'PKG_BUILD_DIR',
                'PKG_BUILD_OUTPUT_DIR',
                'PKG_CACHE_DIR',
                'PKG_CACHE_FILE',
                'PKG_DEFDIR',
                'PKG_NAME',
                'PKG_REVISION',
                'PKG_SITE',
                'PKG_VERSION',
            ]
            self.assertTrue(all(x in data for x in expected_pkg_env_keys))

            # verify that all unexpected environment variables are not set
            unexpected_pkg_env_keys = [
                'PKG_INTERNAL',
            ]
            self.assertTrue(all(x not in data for x in unexpected_pkg_env_keys))

            # sanity check the contents of various environment variables
            self.assertEqual(data['PKG_BUILD_DIR'], data['PKG_BUILD_BASE_DIR'])
            self.assertEqual(data['PKG_NAME'], pkg_name)
            self.assertEqual(data['PKG_REVISION'], 'test-revision')
            self.assertEqual(data['PKG_SITE'], 'mocked-site')
            self.assertEqual(data['PKG_VERSION'], 'test-version')

    def test_pkg_pipeline_pkgenv_internal(self):
        pkg_name = 'test-internal'
        with prepare_testenv(template='pkg-env') as engine:
            pkgs = engine.pkgman.load([pkg_name])

            pipeline = RelengPackagePipeline(engine, engine.opts, {})
            for pkg in pkgs:
                processed = pipeline.process(pkg)
                self.assertTrue(processed)

            # verify package-specific environment variables were set
            self.assertTrue('TEST_INTERNAL_BUILD_OUTPUT_DIR' in os.environ)
            test_outdir = os.environ['TEST_INTERNAL_BUILD_OUTPUT_DIR']
            results = os.path.join(test_outdir, 'invoke-env.json')
            self.assertTrue(os.path.exists(results))

            with open(results, 'r') as f:
                data = json.load(f)

            # ensure that the internal environment flag is set when processing
            # an internal-flagged package
            self.assertTrue('PKG_INTERNAL' in data)

    def test_pkg_pipeline_pkgenv_subdir(self):
        pkg_name = 'test-subdir'
        with prepare_testenv(template='pkg-env') as engine:
            pkgs = engine.pkgman.load([pkg_name])

            pipeline = RelengPackagePipeline(engine, engine.opts, {})
            for pkg in pkgs:
                processed = pipeline.process(pkg)
                self.assertTrue(processed)

            # verify package-specific environment variables were set
            self.assertTrue('TEST_SUBDIR_BUILD_OUTPUT_DIR' in os.environ)
            test_outdir = os.environ['TEST_SUBDIR_BUILD_OUTPUT_DIR']
            results = os.path.join(test_outdir, 'invoke-env.json')
            self.assertTrue(os.path.exists(results))

            with open(results, 'r') as f:
                data = json.load(f)

            # sanity check that the base and build directories are different
            # since a sub-directory has been configured
            build_base_dir = data['PKG_BUILD_BASE_DIR']
            build_dir = data['PKG_BUILD_DIR']
            self.assertNotEqual(build_base_dir, build_dir)
            self.assertNotEqual(build_base_dir, build_dir)
            self.assertTrue(os.path.commonprefix((build_base_dir, build_dir)))



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

import pprint
import os
import unittest


class EnvironmentTestCase(unittest.TestCase):
    """
    environment-managed unit test base

    Provides a base class for unit testing which clears the running environment
    back to its original state after each test run.
    """

    def run(self, result=None):
        """
        run the test

        Run the test, collecting the result into the TestResult object passed as
        result. See `unittest.TestCase.run()` for more details.

        Args:
            result (optional): the test result to populate
        """

        old_env = dict(os.environ)
        try:
            super(EnvironmentTestCase, self).run(result)
        finally:
            os.environ.clear()
            os.environ.update(old_env)

    def dumpenv(self):
        """
        dump the active environment to the standard output stream
        """

        print('-------------------------------')
        pprint.pprint(dict(os.environ))
        print('-------------------------------')



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool

#: executable used to run cvs commands
CVS_COMMAND = 'cvs'

#: list of environment keys to filter from a environment dictionary
CVS_SANITIZE_ENV_KEYS = [
    'CVSIGNORE',
    'CVSREAD',
    'CVSUMASK',
    'CVSWRAPPERS',
    'CVS_SERVER',
]

#: dictionary of environment entries append to the environment dictionary
CVS_EXTEND_ENV = {
    # assume ssh authentication if configured with an :ext: cvsroot
    'CVS_RSH': 'ssh',
}

#: cvs host tool helper
CVS = RelengTool(CVS_COMMAND,
    env_sanitize=CVS_SANITIZE_ENV_KEYS, env_include=CVS_EXTEND_ENV)



# -*- coding: utf-8 -*-
# Copyright 2021-2022 releng-tool

"""
base exception for all custom releng-tool exceptions
"""
class RelengToolException(Exception):
    pass

"""
exception thrown when missing a project's configuration file
"""
class RelengToolMissingConfigurationError(RelengToolException):
    def __init__(self, path):
        RelengToolException.__init__(self, '''\
missing configuration file

The configuration file cannot be found. Ensure the configuration file exists
in the working directory or the provided root directory:

    {}
'''.format(path))

"""
exception thrown when a missing a command for a package's exec call
"""
class RelengToolMissingExecCommand(RelengToolException):
    def __init__(self, pkg):
        RelengToolException.__init__(self, '''\
missing package command

A request has been made to execute a command for a package; however, no command
has been provided. Ensure after specifying an exec call that the following
argument defines the command to be executed.

    releng-tool {}-exec "mycmd arg1 arg2"
'''.format(pkg))

"""
exception thrown when a project's configuration does not provide any packages
"""
class RelengToolMissingPackagesError(RelengToolException):
    def __init__(self, path, key):
        RelengToolException.__init__(self, '''\
no defined packages

The configuration file does not have any defined packages. Ensure a package
list exists with the name of packages to be part of the releng process:

    {}
        {} = ['liba', 'libb', 'libc']
'''.format(path, key))



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.autoreconf import AUTORECONF
from releng_tool.util.io import execute
from releng_tool.util.io import prepare_arguments
from releng_tool.util.io import prepare_definitions
from releng_tool.util.log import err
from releng_tool.util.log import verbose
from releng_tool.util.string import expand

def configure(opts):
    """
    support configuration for autotools projects

    With provided configuration options (``RelengConfigureOptions``), the
    configuration stage will be processed.

    Args:
        opts: configuration options

    Returns:
        ``True`` if the configuration stage is completed; ``False`` otherwise
    """

    # check if autoreconf
    if opts._autotools_autoreconf:
        verbose('configured to run autoreconf')
        if not AUTORECONF.exists():
            err('unable to configure package; autoreconf is not installed')
            return False

        if not AUTORECONF.execute(['--verbose']):
            err('failed to prepare autotools project (autoreconf): {}',
                opts.name)
            return False

    # definitions
    autotools_defs = {
        '--prefix': opts.prefix,
        '--exec-prefix': opts.prefix,
    }
    if opts.conf_defs:
        autotools_defs.update(expand(opts.conf_defs))

    # default options
    autotools_opts = {
    }
    if opts.conf_opts:
        autotools_opts.update(expand(opts.conf_opts))

    # argument building
    autotools_args = [
    ]
    autotools_args.extend(prepare_definitions(autotools_defs))
    autotools_args.extend(prepare_arguments(autotools_opts))

    if not execute(['./configure'] + autotools_args,
            env_update=expand(opts.conf_env), critical=False):
        err('failed to prepare autotools project (configure): {}', opts.name)
        return False

    return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool
#
# Provides an invoke testing executable, which will capture arguments and the
# working directory of a call and output it into a JSON file. Unit tests may
# use this utility file to help verify executions in certain scenarios.

import json
import os
import posixpath
import sys

data = {
    'args': sys.argv,
    'cwd': os.getcwd().replace(os.sep, posixpath.sep),
}

with open('invoke-results.json', 'w') as f:
    json.dump(data, f)



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from releng_tool.defs import GlobalAction
from releng_tool.defs import PkgAction
import multiprocessing
import os

# default directory/file paths
DEFAULT_BUILD_DIR    = 'build'     #: default build container directory
DEFAULT_CACHE_DIR    = 'cache'     #: default cache container directory
DEFAULT_DL_DIR       = 'dl'        #: default download container directory
DEFAULT_HOST_DIR     = 'host'      #: default host container directory
DEFAULT_IMAGES_DIR   = 'images'    #: default images container directory
DEFAULT_LICENSE_DIR  = 'licenses'  #: default licenses container directory
DEFAULT_OUTPUT_DIR   = 'output'    #: default output container directory
DEFAULT_PKG_DIR      = 'package'   #: default package container directory
DEFAULT_STAGING_DIR  = 'staging'   #: default staging container directory
DEFAULT_SYMBOLS_DIR  = 'symbols'   #: default symbols container directory
DEFAULT_TARGET_DIR   = 'target'    #: default target container directory

# default directory/file paths
RELENG_CONF_EXTENDED_NAME = '.releng-tool' #: extended configuration script
RELENG_CONF_NAME = 'releng' #: configuration script filename
RELENG_CONF_OVERRIDES_NAME = 'releng-overrides' #: conf. overrides filename
RELENG_POST_BUILD_NAME = 'releng-post-build' #: post build script filename
FF_PREFIX = '.releng-flag-' #: prefix for all file flags
FF_DEVMODE_NAME = 'devmode' #: postfix for development mode file flag
FF_LOCALSRCS_NAME = 'local-sources' #: postfix for local sources mode file flag

# default other
DEFAULT_SYSROOT_PREFIX = os.sep + 'usr' #: default system root prefix

class RelengEngineOptions:
    """
    engine options

    Configuration options to be passed into an engine instance.

    Args:
        args (optional): handle user-provided configuration options (argparse)
            and apply them to respective attributes
        forward_args (optional): handle arguments provided by the user which can
            be forwarded to the releng-tool project's configuration

    Attributes:
        assets_dir: directory container for cache/download directories
        build_dir: directory container for all builds
        cache_dir: directory container for cache (vcs bare sources)
        cache_ext_transform: transform for cache extension from site path
        conf_point: releng project's configuration
        conf_point_overrides: releng project's configuration overrides
        debug: whether or not debug messages are shown
        default_internal_pkgs: whether or not packages are implicitly internal
        default_pkg_dir: default package directory
        devmode: whether or not development mode is enabled
        dl_dir: directory container for download (archives)
        extern_pkg_dirs: external package directories (if any)
        extract_override: dictionary to override extraction commands
        ff_devmode: the file flag path for development mode detection
        ff_local_srcs: the file flag path for local sources mode detection
        force: whether or not the force flag is set
        forward_args: command line arguments forwarded to configuration script
        gbl_action: the specific global-action to perform (if any)
        host_dir: directory container for host tools
        images_dir: directory container for (final) images
        injected_kv: injected key-value entries for script/env context
        jobs: number of calculated jobs to allow at a given time
        jobsconf: number of jobs to allow at a given time (0: auto)
        license_dir: directory container for license information
        license_header: header content for a generated license file (if any)
        local_srcs: whether or not local sources are used
        no_color_out: whether or not colored messages are shown
        out_dir: directory container for all output data
        pkg_action: the specific package-action to perform (if any)
        post_build_point: resource to process a releng project's post-build work
        prerequisites: list of required host tools (if any)
        quirks: advanced configuration quirks for the running instance
        revision_override: dictionary to override revision values
        root_dir: directory container for all (configuration, output, etc.)
        sites_override: dictionary to override site values
        staging_dir: directory container for staged content
        symbols_dir: directory container for symbols content
        sysroot_prefix: system root prefix
        target_action: the specific package to work on (if any)
        target_action_exec: package-specific executable command (if any)
        target_dir: directory container for target content
        url_mirror: mirror base site for url fetches
        urlopen_context: context to apply for all url open calls
        verbose: whether or not verbose messages are shown
    """
    def __init__(self, args=None, forward_args=None):
        self.assets_dir = None
        self.build_dir = None
        self.cache_dir = None
        self.cache_ext_transform = None
        self.conf_point = None
        self.conf_point_overrides = None
        self.debug = False
        self.default_internal_pkgs = False
        self.default_pkg_dir = None
        self.devmode = None
        self.dl_dir = None
        self.extern_pkg_dirs = []
        self.extract_override = None
        self.ff_devmode = None
        self.ff_local_srcs = None
        self.force = False
        self.forward_args = forward_args
        self.gbl_action = None
        self.host_dir = None
        self.images_dir = None
        self.injected_kv = {}
        self.jobs = 1
        self.jobsconf = 0
        self.license_dir = None
        self.license_header = None
        self.local_srcs = None
        self.no_color_out = False
        self.out_dir = None
        self.pkg_action = None
        self.post_build_point = None
        self.prerequisites = []
        self.quirks = []
        self.revision_override = None
        self.root_dir = None
        self.sites_override = None
        self.staging_dir = None
        self.symbols_dir = None
        self.sysroot_prefix = DEFAULT_SYSROOT_PREFIX
        self.target_action = None
        self.target_action_exec = None
        self.target_dir = None
        self.url_mirror = None
        self.urlopen_context = None
        self.verbose = False

        if args:
            self._handle_arguments(args)
        self._handle_environment_opts()
        self._finalize_options()

    def _handle_arguments(self, args):
        """
        handle argparse-processed arguments to populate engine options

        Accepts the result of an argparse's parsed arguments and updates
        engine options with respective argument options.

        Args:
            args: the arguments
        """
        if args.assets_dir:
            self.assets_dir = os.path.abspath(args.assets_dir)
        if args.cache_dir:
            self.cache_dir = os.path.abspath(args.cache_dir)
        if args.dl_dir:
            self.dl_dir = os.path.abspath(args.dl_dir)
        if args.out_dir:
            self.out_dir = os.path.abspath(args.out_dir)
        if args.root_dir:
            self.root_dir = os.path.abspath(args.root_dir)

        self.debug = args.debug
        self.force = args.force
        self.jobs = self.jobsconf = (args.jobs or 0)
        self.no_color_out = args.nocolorout
        self.verbose = args.verbose

        if args.development:
            self.devmode = True
        if args.injected_kv:
            self.injected_kv = args.injected_kv
        if args.local_sources:
            self.local_srcs = True
        if args.quirk:
            self.quirks.extend(args.quirk)

        if args.action:
            action_val = args.action.lower().replace('-', '_')
            if action_val in GlobalAction:
                self.gbl_action = action_val
            else:
                for subaction_val in PkgAction:
                    if action_val.endswith('_' + subaction_val):
                        self.pkg_action = subaction_val
                        idx = action_val.rindex(subaction_val) - 1
                        self.target_action = args.action[:idx]
                        break

                if subaction_val == PkgAction.EXEC:
                    self.target_action_exec = args.action_exec
                elif not self.pkg_action:
                    self.target_action = args.action

    def _handle_environment_opts(self):
        """
        handle environment variables to populate engine options

        Configure various options which support assignment from an
        environment variable.
        """

        if not self.assets_dir:
            self.assets_dir = os.environ.get('RELENG_ASSETS_DIR')
        if not self.cache_dir:
            self.cache_dir = os.environ.get('RELENG_CACHE_DIR')
        if not self.dl_dir:
            self.dl_dir = os.environ.get('RELENG_DL_DIR')

    def _finalize_options(self):
        """
        finalize all engine options for use

        Ensures all options are properly configured to expected values to ensure
        default values are set which may depend on other currently provided
        options.
        """
        if not self.root_dir:
            self.root_dir = os.getcwd()

        join = os.path.join
        root = self.root_dir

        # asset container
        if self.assets_dir:
            if not self.cache_dir:
                self.cache_dir = join(self.assets_dir, DEFAULT_CACHE_DIR)
            if not self.dl_dir:
                self.dl_dir = join(self.assets_dir, DEFAULT_DL_DIR)
        # root container
        if not self.cache_dir:
            self.cache_dir = join(root, DEFAULT_CACHE_DIR)
        if not self.default_pkg_dir:
            self.default_pkg_dir = join(root, DEFAULT_PKG_DIR)
        if not self.dl_dir:
            self.dl_dir = join(root, DEFAULT_DL_DIR)
        if not self.out_dir:
            self.out_dir = join(root, DEFAULT_OUTPUT_DIR)
        # output container
        if not self.build_dir:
            self.build_dir = join(self.out_dir, DEFAULT_BUILD_DIR)
        if not self.host_dir:
            self.host_dir = join(self.out_dir, DEFAULT_HOST_DIR)
        if not self.images_dir:
            self.images_dir = join(self.out_dir, DEFAULT_IMAGES_DIR)
        if not self.license_dir:
            self.license_dir = join(self.out_dir, DEFAULT_LICENSE_DIR)
        if not self.staging_dir:
            self.staging_dir = join(self.out_dir, DEFAULT_STAGING_DIR)
        if not self.symbols_dir:
            self.symbols_dir = join(self.out_dir, DEFAULT_SYMBOLS_DIR)
        if not self.target_dir:
            self.target_dir = join(self.out_dir, DEFAULT_TARGET_DIR)
        # files
        if not self.conf_point:
            self.conf_point = join(root, RELENG_CONF_NAME)
        if not self.conf_point_overrides:
            self.conf_point_overrides = join(root, RELENG_CONF_OVERRIDES_NAME)
        if not self.ff_devmode:
            self.ff_devmode = join(root, FF_PREFIX + FF_DEVMODE_NAME)
        if not self.ff_local_srcs:
            self.ff_local_srcs = join(root, FF_PREFIX + FF_LOCALSRCS_NAME)
        if not self.post_build_point:
            self.post_build_point = join(root, RELENG_POST_BUILD_NAME)

        # provided fixed job count with auto-configuration (value: 0)
        #
        # The option ``jobsconf`` defines the number of jobs the user wants to
        # use for the process. When providing this information to tools, a value
        # of zero allows a tool to configure its own automatic parallel
        # processing. In some scenarios, a tool may not provide the ability to
        # automatically determine job count to use. For these cases, the value
        # ``jobs`` can be used which either be a value matching ``jobsconf`` if
        # non-zero, or the value will be a non-zero value matching the CPU count
        # of the running system.
        if not self.jobs:
            try:
                # if `sched_getaffinity` is available, use the call the acquire
                # the number of physical cores on the system
                self.jobs = len(os.sched_getaffinity(0))
            except AttributeError:
                # if we cannot guarantee the number of physical cores, make an
                # assumption that the physical core count is half of the
                # (possible logical) core count provided by `cpu_count`
                cpu_count = multiprocessing.cpu_count()
                fuzzy_phy_cores = max(int(cpu_count / 2), 1)
                self.jobs = fuzzy_phy_cores



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.bzr import BZR
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.log import err
from releng_tool.util.log import log
from releng_tool.util.log import note
import os
import sys

def fetch(opts):
    """
    support fetching from bzr sources

    With provided fetch options (``RelengFetchOptions``), the fetch stage will
    be processed.

    Args:
        opts: fetch options

    Returns:
        ``True`` if the fetch stage is completed; ``False`` otherwise
    """

    assert opts
    cache_file = opts.cache_file
    name = opts.name
    revision = opts.revision
    site = opts.site

    if not BZR.exists():
        err('unable to fetch package; bzr is not installed')
        return None

    note('fetching {}...', name)
    sys.stdout.flush()

    cache_dir = os.path.abspath(os.path.join(cache_file, os.pardir))
    if not ensure_dir_exists(cache_dir):
        return None

    log('exporting sources')
    if not BZR.execute(['export', cache_file, site,
            '--format=tgz', '--root=' + name, '--revision=' + revision],
            poll=True):
        err('unable to export module')
        return None

    return cache_file



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.tool import RelengTool

#: executable used to run rsync commands
RSYNC_COMMAND = 'rsync'

#: list of environment keys to filter from a environment dictionary
RSYNC_SANITIZE_ENV_KEYS = [
    'RSYNC_CHECKSUM_LIST',
    'RSYNC_PARTIAL_DIR',
]

#: rsync host tool helper
RSYNC = RelengTool(RSYNC_COMMAND, env_sanitize=RSYNC_SANITIZE_ENV_KEYS)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool

#: executable used to run patch commands
PATCH_COMMAND = 'patch'

#: list of environment keys to filter from a environment dictionary
PATCH_SANITIZE_ENV_KEYS = [
    'PATCH_GET',
    'PATCH_VERSION_CONTROL',
    'POSIXLY_CORRECT',
    'QUOTING_STYLE',
    'SIMPLE_BACKUP_SUFFIX',
    'VERSION_CONTROL',
]

#: patch host tool helper
PATCH = RelengTool(PATCH_COMMAND, env_sanitize=PATCH_SANITIZE_ENV_KEYS)



# -*- coding: utf-8 -*-
# Copyright 2018 releng-tool

from releng_tool.util.string import expand
from releng_tool.util.string import interpret_strings
import os
import unittest

class TestUtilStrings(unittest.TestCase):
    def test_utilstr_expand(self):
        def assertExpand(self, obj, result, kv=None):
            self.assertEqual(expand(obj, kv), result)

        val = expand(None)
        self.assertIsNone(val)

        assertExpand(self, '', '')
        assertExpand(self,
            'this is a simple message', 'this is a simple message')

        os.environ['__RELENGTEST'] = 'test'
        assertExpand(self, '$__RELENGTEST', 'test')
        assertExpand(self, 'A $__RELENGTEST Z', 'A test Z')
        assertExpand(self, 'A $__RELENGTEST$__RELENGTEST Z', 'A testtest Z')
        assertExpand(self, 'longer $__RELENGTEST string', 'longer test string')
        assertExpand(self, 'a/$__RELENGTEST/b', 'a/test/b')
        assertExpand(self, 'a-$__RELENGTEST-b', 'a-test-b')
        assertExpand(self, ' $__RELENGTEST ', ' test ')
        assertExpand(self, '${__RELENGTEST}', 'test')
        assertExpand(self, 'A${__RELENGTEST}Z', 'AtestZ')
        assertExpand(self, 'A${__RELENGTEST}${__RELENGTEST}Z', 'AtesttestZ')
        assertExpand(self,
            'longer ${__RELENGTEST} string', 'longer test string')
        assertExpand(self, ' ${__RELENGTEST} ', ' test ')
        assertExpand(self, '${invalid', '${invalid')
        assertExpand(self, '${}ignored', 'ignored')
        assertExpand(self, '$$escaped', '$escaped')
        assertExpand(self, '${__RELENGTEST}', 'override',
            kv={'__RELENGTEST': 'override'})
        assertExpand(self, ['a', 'b', 'c'], ['a', 'b', 'c'])
        assertExpand(self,
            ['${__RELENGTEST}', 'b', '${__RELENGTEST}'], ['test', 'b', 'test'])
        assertExpand(self, {'a', 'b', 'c'}, {'a', 'b', 'c'})
        assertExpand(self, {'a', '${__RELENGTEST}', 'c'}, {'a', 'test', 'c'})
        assertExpand(self, {'key': 'value'}, {'key': 'value'})
        assertExpand(self,
            {'${__RELENGTEST}': '${__RELENGTEST}'}, {'test': 'test'})

        os.environ.pop('__RELENGTEST', None)
        assertExpand(self, '$__RELENGTEST', '')

    def test_utilstr_interpretstrings(self):
        val = interpret_strings(None)
        self.assertIsNone(val)

        # string returns a list with the single string
        val = interpret_strings('this is a string')
        self.assertEqual(val, ['this is a string'])

        # list returns same list
        val = interpret_strings(['a', 'b'])
        self.assertEqual(val, ['a', 'b'])

        # set returns same set
        val = interpret_strings(('a', 'b'))
        self.assertEqual(val, ('a', 'b'))

        # empty list returns same list
        val = interpret_strings([])
        self.assertEqual(val, [])

        # empty set returns same set
        val = interpret_strings(())
        self.assertEqual(val, ())

        # bad entry returns none
        val = interpret_strings(['a', None])
        self.assertIsNone(val)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.api import RelengExtractOptions
from releng_tool.defs import VcsType
from releng_tool.extract.archive import extract as extract_archive
from releng_tool.extract.git import extract as extract_git
from releng_tool.extract.mercurial import extract as extract_mercurial
from releng_tool.util.api import replicate_package_attribs
from releng_tool.util.hash import HashResult
from releng_tool.util.hash import verify as verify_hashes
from releng_tool.util.io import generate_temp_dir
from releng_tool.util.io import interim_working_dir
from releng_tool.util.io import path_remove
from releng_tool.util.log import debug
from releng_tool.util.log import err
from releng_tool.util.log import note
from releng_tool.util.log import verbose
from releng_tool.util.log import warn
import os
import shutil
import sys

def stage(engine, pkg):
    """
    handles the extraction stage for a package

    With a provided engine and package instance, the extraction stage will be
    processed.

    Args:
        engine: the engine
        pkg: the package being extracted

    Returns:
        ``True`` if the extraction stage is completed; ``False`` otherwise
    """

    # internal packages in local sources mode do not have an extraction stage
    if pkg.is_internal and engine.opts.local_srcs:
        return True

    # skip packages flagged not to extract
    if pkg.no_extraction:
        return True

    note('extracting {}...', pkg.name)
    sys.stdout.flush()

    extract_opts = RelengExtractOptions()
    replicate_package_attribs(extract_opts, pkg)
    extract_opts.cache_dir = pkg.cache_dir
    extract_opts.cache_file = pkg.cache_file
    extract_opts.ext = pkg.ext_modifiers
    extract_opts.name = pkg.name
    extract_opts.revision = pkg.revision
    extract_opts.strip_count = pkg.strip_count
    extract_opts.version = pkg.version
    extract_opts._extract_override = engine.opts.extract_override
    extract_opts._quirks = engine.opts.quirks

    if os.path.exists(pkg.build_dir):
        warn('build directory exists before extraction; removing')

        if not path_remove(pkg.build_dir):
            err('unable to cleanup build directory: ' + pkg.build_dir)
            return False

    # prepare and step into the a newly created working directory
    #
    # An extractor will take the contents of an archive, cache directory or
    # other fetched content and populate the "work" directory. On successful
    # extraction (or moving resources), the work directory will be moved to the
    # package's respective build directory.
    out_dir = engine.opts.out_dir
    with generate_temp_dir(out_dir) as work_dir:
        with interim_working_dir(work_dir):
            extract_opts.work_dir = work_dir

            extracter = None
            hash_exclude = []
            extract_types = engine.registry.extract_types
            if pkg.extract_type and pkg.extract_type in extract_types:
                def _(opts):
                    return engine.registry.extract_types[pkg.vcs_type].extract(
                        pkg.vcs_type, opts)
                extracter = _
            elif pkg.vcs_type in extract_types:
                extracter = extract_types[pkg.vcs_type].extract
            elif pkg.vcs_type == VcsType.GIT:
                extracter = extract_git
            elif pkg.vcs_type == VcsType.HG:
                extracter = extract_mercurial
            elif os.path.isfile(pkg.cache_file):
                cache_basename = os.path.basename(pkg.cache_file)
                hash_exclude.append(cache_basename)
                extracter = extract_archive

            if not extracter:
                err('extract type is not implemented: {}', pkg.vcs_type)
                return False

            # perform the extract request
            extracted = extracter(extract_opts)
            if not extracted:
                return False

            result = verify_hashes(pkg.hash_file, work_dir, hash_exclude)
            if result == HashResult.VERIFIED:
                pass
            elif result == HashResult.BAD_PATH:
                if not pkg.is_internal:
                    warn('missing hash file for package: ' + pkg.name)
            elif result == HashResult.EMPTY:
                if not pkg.is_internal:
                    verbose('hash file for package is empty: ' + pkg.name)
            elif result in (HashResult.BAD_FORMAT, HashResult.MISMATCH,
                    HashResult.MISSING_LISTED, HashResult.UNSUPPORTED):
                return False
            else:
                err('invalid extract operation (internal error; '
                    'hash-check failure: {})', result)
                return False

        debug('extraction successful; moving sources into package output '
            'directory: ' + pkg.build_dir)
        shutil.move(work_dir, pkg.build_dir)

    return True



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.util.log import err
import sys

def platform_exit(msg=None, code=None):
    """
    exit out of the releng-tool process

    Provides a convenience method to help invoke a system exit call without
    needing to explicitly use ``sys``. A caller can provide a message to
    indicate the reason for the exit. The provide message will output to
    standard error. The exit code, if not explicit set, will vary on other
    arguments. If a message is provided to this call, the default exit code will
    be ``1``. If no message is provided, the default exit code will be ``0``.
    In any case, if the caller explicitly sets a code value, the provided code
    value will be used.

    An example when using in the context of script helpers is as follows:

    .. code-block:: python

        releng_exit('there was an error performing this task')

    Args:
        msg (optional): error message to print
        code (optional): exit code; defaults to 0 if no message or defaults to 1
            if a message is set

    Raises:
        SystemExit: always raised
    """

    if msg:
        err(msg)
        if code is None:
            code = 1
    elif code is None:
        code = 0
    sys.exit(code)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.scp import SCP
from releng_tool.util.log import err
from releng_tool.util.log import log
from releng_tool.util.log import note
import sys

def fetch(opts):
    """
    support fetching from scp sources

    With provided fetch options (``RelengFetchOptions``), the fetch stage will
    be processed.

    Args:
        opts: fetch options

    Returns:
        ``True`` if the fetch stage is completed; ``False`` otherwise
    """

    assert opts
    cache_file = opts.cache_file
    name = opts.name
    site = opts.site
    work_dir = opts.work_dir

    if not SCP.exists():
        err('unable to fetch package; scp is not installed')
        return None

    note('fetching {}...', name)
    sys.stdout.flush()

    if not SCP.execute(['-o', 'BatchMode yes', site, cache_file], cwd=work_dir):
        err('unable to secure-copied file from target')
        return None
    log('successfully secure-copied file from target')

    return cache_file



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from releng_tool.util.enum import Enum

class ConfKey(Enum):
    """
    configuration file keys

    Defines a series of attributes which define every support project
    configuration key supported by this tool. Project configuration keys
    are in a lowercase format.

    Attributes:
        CACHE_EXT_TRANSFORM: cache extension transform
        DEFINTERN: packages are internal (implicitly)
        EXTENSIONS: project releng-extension list
        EXTEN_PKGS: project external packages list
        LICENSE_HEADER: license header information
        OVERRIDE_REV: revision overriding dictionary
        OVERRIDE_SITES: site overriding dictionary
        OVERRIDE_TOOLS: extract-tool overriding
        PKGS: project's package (name) list
        PREREQUISITES: project's host-tool prerequisites
        QUIRKS: configure quirks to apply
        SYSROOT_PREFIX: project's default sys-root prefix
        URL_MIRROR: mirror base site for url fetches
        URLOPEN_CONTEXT: context to use for urlopen
    """
    CACHE_EXT_TRANSFORM = 'cache_ext'
    DEFINTERN = 'default_internal'
    EXTENSIONS = 'extensions'
    EXTEN_PKGS = 'external_packages'
    LICENSE_HEADER = 'license_header'
    OVERRIDE_REV = 'override_revisions'
    OVERRIDE_SITES = 'override_sites'
    OVERRIDE_TOOLS = 'override_extract_tools'
    PKGS = 'packages'
    PREREQUISITES = 'prerequisites'
    QUIRKS = 'quirks'
    SYSROOT_PREFIX = 'sysroot_prefix'
    URL_MIRROR = 'url_mirror'
    URLOPEN_CONTEXT = 'urlopen_context'

class Rpk(Enum):
    """
    releng package keys (postfixes)

    Defines a series of attributes which define every support package
    configuration key supported by this tool. Package configuration keys
    are in an uppercase format.

    Attributes:
        BUILD_SUBDIR: sub-directory in fetched to find root src
        DEPS: list of package dependencies
        DEVMODE_IGNORE_CACHE: whether or not ignore cache
        DEVMODE_REVISION: devmode-rev to acquire from srcs
        EXTENSION: filename extension for package (if needed)
        EXTERNAL: whether or not package is considered "external"
        EXTOPT: extension-defined package modifiers (if any)
        EXTRACT_TYPE: extraction type for sources
        FETCH_OPTS: fetch options (if any)
        FIXED_JOBS: fixed job count for the project
        GIT_CONFIG: git configurations to set (if any)
        GIT_DEPTH: git fetch depth (if any)
        GIT_REFSPECS: additional git refspecs to fetch (if any)
        GIT_SUBMODULES: fetch any submodules (if any)
        GIT_VERIFY_REVISION: verify signed revisions
        INSTALL_TYPE: install container target for the package
        INTERNAL: whether or not package is considered "internal"
        LICENSE: license information for the package
        LICENSE_FILES: source file(s) with license information
        NO_EXTRACTION: whether or not package extraction is done
        PREFIX: system root prefix override (if needed)
        REVISION: revision to acquire from sources (if any)
        SITE: site where to fetch package sources
        SKIP_REMOTE_CONFIG: skip any remote configuration
        SKIP_REMOTE_SCRIPTS: skip any remote scripts
        STRIP_COUNT: strip count for archive extract
        TYPE: type of project the package is
        VCS_TYPE: type of project the package's fetch source is
        VERSION: the version of the package
        # (package type - common)
        CONF_DEFS: package-type configuration definitions
        CONF_ENV: package-type configuration environment values
        CONF_OPTS: package-type configuration options
        BUILD_DEFS: package-type build definitions
        BUILD_ENV: package-type build environment values
        BUILD_OPTS: package-type build options
        INSTALL_DEFS: package-type install definitions
        INSTALL_ENV: package-type install environment values
        INSTALL_OPTS: package-type install options
        # (package type - autotools)
        AUTOTOOLS_AUTORECONF: autotools /w autoreconf
        # (package type - cmake)
        CMAKE_NOINSTALL: skip cmake install stage
        # (package type - python)
        PYTHON_INTERPRETER: python interpreter
    """
    BUILD_SUBDIR = 'BUILD_SUBDIR'
    DEPS = 'DEPENDENCIES'
    DEVMODE_IGNORE_CACHE = 'DEVMODE_IGNORE_CACHE'
    DEVMODE_REVISION = 'DEVMODE_REVISION'
    EXTENSION = 'EXTENSION'
    EXTERNAL = 'EXTERNAL'
    EXTOPT = 'EXTOPT'
    EXTRACT_TYPE = 'EXTRACT_TYPE'
    FETCH_OPTS = 'FETCH_OPTS'
    FIXED_JOBS = 'FIXED_JOBS'
    GIT_CONFIG = 'GIT_CONFIG'
    GIT_DEPTH = 'GIT_DEPTH'
    GIT_REFSPECS = 'GIT_REFSPECS'
    GIT_SUBMODULES = 'GIT_SUBMODULES'
    GIT_VERIFY_REVISION = 'GIT_VERIFY_REVISION'
    INSTALL_TYPE = 'INSTALL_TYPE'
    INTERNAL = 'INTERNAL'
    LICENSE = 'LICENSE'
    LICENSE_FILES = 'LICENSE_FILES'
    NO_EXTRACTION = 'NO_EXTRACTION'
    PREFIX = 'PREFIX'
    REVISION = 'REVISION'
    SITE = 'SITE'
    SKIP_REMOTE_CONFIG = 'SKIP_REMOTE_CONFIG'
    SKIP_REMOTE_SCRIPTS = 'SKIP_REMOTE_SCRIPTS'
    STRIP_COUNT = 'STRIP_COUNT'
    TYPE = 'TYPE'
    VCS_TYPE = 'VCS_TYPE'
    VERSION = 'VERSION'
    # (package type - common)
    CONF_DEFS = 'CONF_DEFS'
    CONF_ENV = 'CONF_ENV'
    CONF_OPTS = 'CONF_OPTS'
    BUILD_DEFS = 'BUILD_DEFS'
    BUILD_ENV = 'BUILD_ENV'
    BUILD_OPTS = 'BUILD_OPTS'
    INSTALL_DEFS = 'INSTALL_DEFS'
    INSTALL_ENV = 'INSTALL_ENV'
    INSTALL_OPTS = 'INSTALL_OPTS'
    # (package type - autotools)
    AUTOTOOLS_AUTORECONF = 'AUTOTOOLS_AUTORECONF'
    # (package type - cmake)
    CMAKE_NOINSTALL = 'CMAKE_NOINSTALL'
    # (package type - python)
    PYTHON_INTERPRETER = 'PYTHON_INTERPRETER'

class GlobalAction(Enum):
    """
    specific stage action to perform

    A user can request a (global) action to perform over the default process.
    For example, a user can request to "fetch" and only the fetching stage will
    occur for registered packages.

    Attributes:
        CLEAN: clean the working state
        DISTCLEAN: pristine state clean with cache/dl clear
        EXTRACT: process all packages through extraction stage
        FETCH: process all packages through fetch stage
        INIT: initialize example structure
        LICENSES: generate license information for a project
        MRPROPER: pristine state clean (e.g. configurations)
        PATCH: process all packages through patch stage
    """
    CLEAN = 'clean'
    DISTCLEAN = 'distclean'
    EXTRACT = 'extract'
    FETCH = 'fetch'
    INIT = 'init'
    LICENSES = 'licenses'
    MRPROPER = 'mrproper'
    PATCH = 'patch'

class PkgAction(Enum):
    """
    package-specific stage action to perform

    A user can request a package action to perform over the default process.
    When a package-specific action is requested, the process will perform all
    dependencies for the target action's stage before completing the target
    stage. For example, a user can request to perform a package's "extract"
    stage, which will result in ensure the package's fetch stage is complete
    (and possibility other package dependencies) performing (and stopping after)
    the extraction stage.

    Attributes:
        BUILD: process a package till end of the build stage
        CLEAN: process a package till end of the clean stage
        CONFIGURE: process a package till end of the configure stage
        DISTCLEAN: pristine state clean state of the package with cache/dl clear
        EXEC: perform an action in the package's directory
        EXTRACT: process a package till end of the extraction stage
        FETCH: process a package till end of the fetch stage
        INSTALL: process a package till end of the install stage
        LICENSE: generate license information for a package
        PATCH: process a package till end of the patch stage
        REBUILD: perform a re-build of a package
        REBUILD_ONLY: perform a re-build of a package and stop
        RECONFIGURE: perform a re-configuration of a package
        RECONFIGURE_ONLY: perform a re-configuration of a package and stop
        REINSTALL: perform a re-install of a package
    """
    BUILD = 'build'
    CLEAN = 'clean'
    CONFIGURE = 'configure'
    DISTCLEAN = 'distclean'
    EXEC = 'exec'
    EXTRACT = 'extract'
    FETCH = 'fetch'
    INSTALL = 'install'
    LICENSE = 'license'
    PATCH = 'patch'
    REBUILD = 'rebuild'
    REBUILD_ONLY = 'rebuild_only'
    RECONFIGURE = 'reconfigure'
    RECONFIGURE_ONLY = 'reconfigure_only'
    REINSTALL = 'reinstall'

class PackageType(Enum):
    """
    package types

    Defines supported package types for deciding which method if configuring,
    building and installing is performed.

    Attributes:
        AUTOTOOLS: autotools-based package
        CMAKE: cmake-based package
        PYTHON: python-based package
        SCRIPT: releng script-based package
    """
    AUTOTOOLS = 'autotools'
    CMAKE = 'cmake'
    PYTHON = 'python'
    SCRIPT = 'script'

class PackageInstallType(Enum):
    """
    package install types

    Defines supported package installation types for deciding which the location
    to push resources during the installation phase.

    Attributes:
        HOST: install to the host container
        IMAGES: install to the images container
        STAGING: install to the staging container
        STAGING_AND_TARGET: install to the staging and target containers
        TARGET: install to the target container
    """
    HOST = 'host'
    IMAGES = 'images'
    STAGING = 'staging'
    STAGING_AND_TARGET = 'staging_and_target'
    TARGET = 'target'

class VcsType(Enum):
    """
    version control system types

    Defines supported version control system types for decided which fetching
    processing is used when acquiring resources.

    Attributes:
        CVS: concurrent versions system
        GIT: git
        HG: mercurial
        LOCAL: no version control (local interim-development package)
        NONE: no version control (placeholder package)
        RSYNC: secure copy
        SCP: secure copy
        SVN: subversion
        URL: url (http, https, ftp, file, etc.)
    """
    BZR = 'bzr'
    CVS = 'cvs'
    GIT = 'git'
    HG = 'hg'
    LOCAL = 'local'
    NONE = 'none'
    RSYNC = 'rsync'
    SCP = 'scp'
    SVN = 'svn'
    URL = 'url'



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.python import PYTHON
from releng_tool.tool.python import PYTHON_EXTEND_ENV
from releng_tool.tool.python import PythonTool
from releng_tool.util.io import prepare_arguments
from releng_tool.util.io import prepare_definitions
from releng_tool.util.log import err
from releng_tool.util.string import expand
import os

def build(opts):
    """
    support building python projects

    With provided build options (``RelengBuildOptions``), the build stage will
    be processed.

    Args:
        opts: build options

    Returns:
        ``True`` if the building stage is completed; ``False`` otherwise
    """

    if opts._python_interpreter:
        python_tool = PythonTool(opts._python_interpreter,
            env_include=PYTHON_EXTEND_ENV)
    else:
        python_tool = PYTHON

    if not python_tool.exists():
        err('unable to build package; python is not installed')
        return False

    # definitions
    python_defs = {
    }
    if opts.build_defs:
        python_defs.update(expand(opts.build_defs))

    # default options
    python_opts = {
    }
    if opts.build_opts:
        python_opts.update(expand(opts.build_opts))

    # default environment
    path1 = python_tool.path(sysroot=opts.staging_dir, prefix=opts.prefix)
    path2 = python_tool.path(sysroot=opts.target_dir, prefix=opts.prefix)
    env = {
        'PYTHONPATH': path1 + os.pathsep + path2
    }

    # apply package-specific environment options
    if opts.build_env:
        env.update(expand(opts.build_env))

    # argument building
    python_args = [
        'setup.py',
        # ignore user's pydistutils.cfg
        '--no-user-cfg',
        # invoke the build operation
        'build',
    ]
    python_args.extend(prepare_definitions(python_defs))
    python_args.extend(prepare_arguments(python_opts))

    if not python_tool.execute(python_args, env=env):
        err('failed to build python project: {}', opts.name)
        return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.cvs import CVS
from releng_tool.util.io import ensure_dir_exists
from releng_tool.util.io import interpret_stem_extension
from releng_tool.util.log import err
from releng_tool.util.log import log
from releng_tool.util.log import note
import os
import sys
import tarfile

def fetch(opts):
    """
    support fetching from cvs sources

    With provided fetch options (``RelengFetchOptions``), the fetch stage will
    be processed.

    Args:
        opts: fetch options

    Returns:
        ``True`` if the fetch stage is completed; ``False`` otherwise
    """

    assert opts
    cache_file = opts.cache_file
    name = opts.name
    revision = opts.revision
    site = opts.site
    work_dir = opts.work_dir

    cache_basename = os.path.basename(cache_file)
    cache_stem, __ = interpret_stem_extension(cache_basename)

    if not CVS.exists():
        err('unable to fetch package; cvs is not installed')
        return None

    note('fetching {}...', name)
    sys.stdout.flush()

    try:
        cvsroot, module = site.rsplit(' ', 1)
    except ValueError:
        err('''\
improper cvs site defined

The provided CVS site does not define both the CVSROOT as well as the target
module to checkout. For example:

    :pserver:anonymous@cvs.example.com:/var/lib/cvsroot mymodule

 Site: {}''', site)
        return None

    log('checking out sources')
    if not CVS.execute(['-d', cvsroot, 'checkout', '-d', cache_stem,
            '-r', revision, module], cwd=work_dir):
        err('unable to checkout module')
        return None

    cvs_module_dir = os.path.join(work_dir, cache_stem)
    if not os.path.exists(cvs_module_dir):
        err('no sources available for the provided revision')
        return None

    log('caching sources')
    def cvs_filter(info):
        if info.name.endswith('CVS'):
            return None
        return info

    cache_dir = os.path.abspath(os.path.join(cache_file, os.pardir))
    if not ensure_dir_exists(cache_dir):
        return None

    with tarfile.open(cache_file, 'w:gz') as tar:
        tar.add(cvs_module_dir, arcname=cache_stem, filter=cvs_filter)

    return cache_file



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from tests.support.site_tool_test import TestSiteToolBase
import os
import sys

class TestToolCmake(TestSiteToolBase):
    @classmethod
    def setUpClass(cls):
        cls.filename = 'testapp'
        if sys.platform == 'win32':
            cls.filename += '.exe'

        cls.prefix = os.path.join('usr', 'bin')

    def prepare_global_action(self):
        return None # use releng-tool default

    def tool_template(self):
        return 'cmake'

    def test_tool_cmake_default(self):
        rv = self.engine.run()
        self.assertTrue(rv)

        bin_dir = os.path.join(self.engine.opts.staging_dir, self.prefix)
        executable = os.path.join(bin_dir, self.filename)
        self.assertFalse(os.path.exists(executable))

        bin_dir = os.path.join(self.engine.opts.target_dir, self.prefix)
        executable = os.path.join(bin_dir, self.filename)
        self.assertTrue(os.path.exists(executable))

    def test_tool_cmake_host(self):
        self.defconfig_add('INSTALL_TYPE', 'host')

        rv = self.engine.run()
        self.assertTrue(rv)

        bin_dir = os.path.join(self.engine.opts.host_dir, self.prefix)
        executable = os.path.join(bin_dir, self.filename)
        self.assertTrue(os.path.exists(executable))

    def test_tool_cmake_staging(self):
        self.defconfig_add('INSTALL_TYPE', 'staging')

        rv = self.engine.run()
        self.assertTrue(rv)

        bin_dir = os.path.join(self.engine.opts.staging_dir, self.prefix)
        executable = os.path.join(bin_dir, self.filename)
        self.assertTrue(os.path.exists(executable))

        bin_dir = os.path.join(self.engine.opts.target_dir, self.prefix)
        executable = os.path.join(bin_dir, self.filename)
        self.assertFalse(os.path.exists(executable))

    def test_tool_cmake_staging_and_target(self):
        self.defconfig_add('INSTALL_TYPE', 'staging_and_target')

        rv = self.engine.run()
        self.assertTrue(rv)

        bin_dir = os.path.join(self.engine.opts.staging_dir, self.prefix)
        executable = os.path.join(bin_dir, self.filename)
        self.assertTrue(os.path.exists(executable))

        bin_dir = os.path.join(self.engine.opts.target_dir, self.prefix)
        executable = os.path.join(bin_dir, self.filename)
        self.assertTrue(os.path.exists(executable))

    def test_tool_cmake_noinstall(self):
        self.defconfig_add('CMAKE_NOINSTALL', True)

        rv = self.engine.run()
        self.assertTrue(rv)

        bin_dir = os.path.join(self.engine.opts.target_dir, self.prefix)
        executable = os.path.join(bin_dir, self.filename)
        self.assertFalse(os.path.exists(executable))



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from releng_tool import __version__ as releng_version
from releng_tool.util.log import err
from runpy import run_path
import inspect
import os
import sys

def releng_include(file_path):
    """
    include/execute a script

    The provided call will execute code at the provided file path. The path
    will be relative to the caller's script, unless an absolute path is
    provided. The executed script will be initialized with globals matching
    the caller's script.

    An example when using in the context of script helpers is as follows:

    .. code-block:: python

        # load "my-other-script" found alongside the current script
        releng_include('my-other-script')

    Args:
        file_path: the script to invoke
    """

    caller_stack = inspect.stack()[1]

    if os.path.isabs(file_path):
        target_script = file_path
    else:
        invoked_script = caller_stack[1]
        invoked_script_base = os.path.dirname(invoked_script)
        target_script = os.path.join(invoked_script_base, file_path)

    globals = caller_stack[0].f_globals
    run_path(target_script, init_globals=globals)

def require_version(version, quiet=False, critical=True):
    """
    perform a required-version check

    Enables a caller to explicitly check for a required releng-tool version.
    Invoking this function with a dotted-separated ``version`` string, the
    string will be parsed and compared with the running releng-tool version.
    If the required version is met, this method will have no effect. In the
    event that the required version is not met, the exception ``SystemExit``
    will be raised if the critical flag is set; otherwise this call will
    return ``False``.

    An example when using in the context of script helpers is as follows:

    .. code-block:: python

        # ensure we are using releng-tool v1
        releng_require_version('1.0.0')

    Args:
        version: dotted-separated version string
        quiet (optional): whether or not to suppress output
        critical (optional): whether or not to stop execution on failure

    Returns:
        ``True`` if the version check is met; ``False`` if the version check
        has failed

    Raises:
        SystemExit: if the version check fails with ``critical=True``
    """

    rv = True

    if version:
        requested = version.split('.')
        current = releng_version.split('.')
        rv = requested <= current
        if not rv:
            if not quiet:
                args = {
                    'detected': releng_version,
                    'required': version,
                }
                err('''
required releng-tool version check has failed

This project has indicated a required minimum version of releng-tool to
be installed on this system; however, an older version has been
detected:

    (required) {required}
    (detected) {detected}

Please update to a more recent version:

 https://docs.releng.io/install/
'''.strip().format(**args))

            if critical:
                sys.exit(-1)

    return rv



# -*- coding: utf-8 -*-
# Copyright 2018-2022 releng-tool

from releng_tool.util.io import opt_file
from releng_tool.util.io import run_script
from releng_tool.util.log import verbose
import os

#: filename of the script to execute the installation operation (if any)
INSTALL_SCRIPT = 'install'

def install(opts):
    """
    support installation project-defined scripts

    With provided installation options (``RelengInstallOptions``), the
    installation stage will be processed.

    Args:
        opts: installation options

    Returns:
        ``True`` if the installation stage is completed; ``False`` otherwise
    """

    assert opts
    build_dir = opts.build_dir
    def_dir = opts.def_dir
    env = opts.env

    install_script_filename = '{}-{}'.format(opts.name, INSTALL_SCRIPT)
    install_script = os.path.join(def_dir, install_script_filename)
    install_script, install_script_exists = opt_file(install_script)
    if not install_script_exists:
        if (opts._skip_remote_scripts or
                'releng.disable_remote_scripts' in opts._quirks):
            return True

        install_script_filename = '{}-{}'.format('releng', INSTALL_SCRIPT)
        install_script = os.path.join(build_dir, install_script_filename)
        install_script, install_script_exists = opt_file(install_script)
        if not install_script_exists:
            return True

    if not run_script(install_script, env, subject='install'):
        return False

    verbose('install script executed: ' + install_script)
    return True



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from tests import prepare_testenv
import os
import unittest
from releng_tool.exceptions import RelengToolMissingExecCommand


class TestEngineExec(unittest.TestCase):
    def test_engine_run_exec_arg_missing(self):
        config = {
            'action': 'test-exec',
        }

        with prepare_testenv(config=config, template='exec') as engine:
            with self.assertRaises(RelengToolMissingExecCommand):
                engine.run()

    def test_engine_run_exec_failed(self):
        config = {
            'action': 'test-exec',
            'action_exec': 'python fail.py',
        }

        with prepare_testenv(config=config, template='exec') as engine:
            rv = engine.run()
            self.assertFalse(rv)

            flag = os.path.join(engine.opts.target_dir, 'invoked-fail')
            self.assertTrue(os.path.exists(flag))

    def test_engine_run_exec_success(self):
        config = {
            'action': 'test-exec',
            'action_exec': 'python success.py',
        }

        with prepare_testenv(config=config, template='exec') as engine:
            rv = engine.run()
            self.assertTrue(rv)

            flag = os.path.join(engine.opts.target_dir, 'invoked-success')
            self.assertTrue(os.path.exists(flag))



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool.python import PYTHON
from releng_tool.tool.python import PYTHON_EXTEND_ENV
from releng_tool.tool.python import PythonTool
from releng_tool.util.io import prepare_arguments
from releng_tool.util.io import prepare_definitions
from releng_tool.util.log import err
from releng_tool.util.string import expand
import sys

def install(opts):
    """
    support installation python projects

    With provided installation options (``RelengInstallOptions``), the
    installation stage will be processed.

    Args:
        opts: installation options

    Returns:
        ``True`` if the installation stage is completed; ``False`` otherwise
    """

    if opts._python_interpreter:
        python_tool = PythonTool(opts._python_interpreter,
            env_include=PYTHON_EXTEND_ENV)
    else:
        python_tool = PYTHON

    if not python_tool.exists():
        err('unable to install package; python is not installed')
        return False

    # definitions
    python_defs = {
        '--prefix': opts.prefix,
    }
    if opts.install_defs:
        python_defs.update(expand(opts.install_defs))

    # always remove the prefix value if:
    #  - *nix: setup.py may ignore provided `--root` value with an "/" prefix
    #  - win32: does not use the prefix value
    if python_defs['--prefix'] == '/' or sys.platform == 'win32':
        del python_defs['--prefix']

    # default options
    python_opts = {
    }
    if opts.install_opts:
        python_opts.update(expand(opts.install_opts))

    # argument building
    python_args = [
        'setup.py',
        # ignore user's pydistutils.cfg
        '--no-user-cfg',
        # invoke the install operation
        'install',
        # avoid building pyc files
        '--no-compile',
    ]
    python_args.extend(prepare_definitions(python_defs))
    python_args.extend(prepare_arguments(python_opts))

    # install to target destination(s)
    #
    # If the package already defines a root path, use it over any other
    # configured destination directories.
    env = expand(opts.install_env)
    if '--root' in python_opts:
        if not python_tool.execute(python_args, env=env):
            err('failed to install python project: {}', opts.name)
            return False
    else:
        # install to each destination
        for dest_dir in opts.dest_dirs:
            if not python_tool.execute(python_args + ['--root', dest_dir],
                    env=env):
                err('failed to install python project: {}', opts.name)
                return False

    return True



# -*- coding: utf-8 -*-
# Copyright 2022 releng-tool

from releng_tool.util.io import ls
from tests import prepare_workdir
from tests import redirect_stdout
import os
import unittest


class TestUtilIoLs(unittest.TestCase):
    def test_utilio_ls_invalid_directory(self):
        with redirect_stdout() as stream:
            listed = ls(os.path.realpath(__file__))

        self.assertFalse(listed)
        self.assertEqual(stream.getvalue(), '')

    def test_utilio_ls_missing_directory(self):
        with redirect_stdout() as stream:
            listed = ls('missing-directory')

        self.assertFalse(listed)
        self.assertEqual(stream.getvalue(), '')

    def test_utilio_ls_valid_directory_contents(self):
        base_dir = os.path.dirname(os.path.realpath(__file__))

        with redirect_stdout() as stream:
            listed = ls(base_dir)

        self.assertTrue(listed)

        entries = stream.getvalue().split('\n')
        self.assertTrue('assets/' in entries)
        self.assertTrue('__init__.py' in entries)
        self.assertTrue('test_util_io_cat.py' in entries)

    def test_utilio_ls_valid_directory_empty(self):
        with redirect_stdout() as stream:
            with prepare_workdir() as empty_dir:
                listed = ls(empty_dir)

        self.assertTrue(listed)
        self.assertEqual(stream.getvalue(), '')



# -*- coding: utf-8 -*-
# Copyright 2021 releng-tool

from releng_tool.util.io import run_script
from tests import prepare_testenv
import os
import sys
import unittest

class TestEngineRunBinPaths(unittest.TestCase):
    def test_engine_run_bin_paths_host_output_custom_prefix(self):
        with prepare_testenv(template='hosts-check-prefix') as engine:
            engine.run()

            expected = 'releng-tool-test-script-custom'
            self._validate_test_script(expected)

    def test_engine_run_bin_paths_host_output_default_prefix(self):
        with prepare_testenv(template='hosts-check') as engine:
            engine.run()

            expected = 'releng-tool-test-script-bin'
            self._validate_test_script(expected)

    def test_engine_run_bin_paths_host_output_root(self):
        with prepare_testenv(template='hosts-check') as engine:
            engine.run()

            expected = 'releng-tool-test-script-host'
            self._validate_test_script(expected)

    def test_engine_run_bin_paths_project_root(self):
        with prepare_testenv(template='hosts-check') as engine:
            engine.run()

            expected = 'releng-tool-test-script-root'
            self._validate_test_script(expected)

    def _validate_test_script(self, name):
        gbls = None

        for path in sys.path:
            script = os.path.join(path, name)

            if os.path.exists(script):
                gbls = run_script(script, gbls, catch=False)
                break

        self.assertIsNotNone(gbls)
        self.assertEqual(gbls['var'], name)



# -*- coding: utf-8 -*-
# Copyright 2018-2021 releng-tool

from releng_tool.tool import RelengTool
from releng_tool.util.compat import which
from releng_tool.util.log import debug

#: executable used to run scp commands
SCP_COMMAND = 'scp'

#: list of environment keys to filter from a environment dictionary
SCP_SANITIZE_ENV_KEYS = [
    # remove the possibility for authenticated prompts
    'SSH_ASKPASS',
]

class ScpTool(RelengTool):
    """
    scp host tool

    Provides addition helper methods for scp-based tool interaction.
    """

    def exists(self):
        """
        return whether or not the host tool exists

        Returns whether or not the tool is available on the host for use.

        Returns:
            ``True``, if the tool exists; ``False`` otherwise
        """
        if self.tool in RelengTool.detected:
            return RelengTool.detected[self.tool]

        if which(self.tool):
            debug('{} tool is detected on this system', self.tool)
            RelengTool.detected[self.tool] = True
        else:
            debug('{} tool is not detected on this system', self.tool)
            RelengTool.detected[self.tool] = False

        return RelengTool.detected[self.tool]

#: scp host tool helper
SCP = ScpTool(SCP_COMMAND, env_sanitize=SCP_SANITIZE_ENV_KEYS)
